# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Two 8-step trigger, gate & CV sequencers based on the binary representation of an 8-bit number
"""

from europi import *
from europi_script import EuroPiScript

import configuration
import time


class BittySequence:
    """
    A container class for one sequencer
    """
    def __init__(self, button_in, trigger_out, gate_out, cv_out, use_gray_encoding=False):
        """
        Create a new sequencer

        @param button_in  The button the user can press to manually advance the sequence
        @param trigger_out  The CV output for the trigger signal
        @param gate_out  The CV output for the gate signal
        @param cv_out  The CV output for the CV signal
        @param use_gray_encoding  If true, we use gray encoding instead of traditional binary
        """
        button_in.handler(self.advance)
        button_in.handler_falling(self.trigger_off)

        self.trigger_out = trigger_out
        self.gate_out = gate_out
        self.cv_out = cv_out

        self.use_gray_encoding = use_gray_encoding

        # the integer representation of our sequence
        # if we're not using gray encoding this should be the same as our binary sequence
        self.sequence_n = 0

        # the raw binary pattern that represents our sequence
        self.binary_sequence = 0x00

        # the current step
        self.step = 0

        # is the current output state in need of refreshing?
        self.output_dirty = False

        # turn everything off initially
        self.trigger_out.off()
        self.gate_out.off()
        self.cv_out.off()

    def advance(self):
        self.step = (self.step + 1) & 0x07  # restrict this to 0-7
        self.output_dirty = True

    def trigger_off(self):
        self.trigger_out.off()

    def apply_output(self):
        now = time.ticks_ms()

        if self.current_bit:
            self.trigger_out.on()
            self.gate_out.on()
        else:
            self.trigger_out.off()
            self.gate_out.off()

        self.cv_out.voltage(europi_config.MAX_OUTPUT_VOLTAGE * self.cv_sequence / 255)

        self.output_dirty = False

    def change_sequence(self, n):
        self.sequence_n = n

        if self.use_gray_encoding:
            # convert the number from traditional binary to its gray encoding equivalent
            n = (n & 0xff) ^ ((n & 0xff) >> 1)
        else:
            n = n & 0xff

        self.binary_sequence = n

    @property
    def shifted_sequence(self):
        return ((self.binary_sequence << self.step) & 0xff) | ((self.binary_sequence & 0xff) >> (8 - self.step))

    @property
    def cv_sequence(self):
        # reverse the bits of the shifted sequence
        s = self.shifted_sequence
        cv = 0x00
        while s:
            cv = cv << 1
            cv = cv | (s & 0x01)
            s = s >> 1
        return cv & 0xff

    @property
    def current_bit(self):
        return (self.shifted_sequence >> 7) & 0x01


class IttyBitty(EuroPiScript):
    def __init__(self):
        super().__init__()

        self.sequencers = [
            BittySequence(b1, cv1, cv2, cv3, use_gray_encoding=self.config.USE_GRAY_ENCODING),
            BittySequence(b2, cv4, cv5, cv6, use_gray_encoding=self.config.USE_GRAY_ENCODING),
        ]

        @din.handler
        def on_clock_rise():
            for s in self.sequencers:
                s.advance()

        @din.handler_falling
        def on_clock_fall():
            for s in self.sequencers:
                s.trigger_off()

    @classmethod
    def config_points(cls):
        return [
            # If true, use gray encoding instead of standard binary
            # Gray encding flips a single bit at each step, meaning any two adjacent
            # sequences differ by only 1 bit
            configuration.boolean(
                "USE_GRAY_ENCODING",
                False
            ),

            # Flags to enable AIN to control channel A and/or channel B
            # when enabled, the knob acts as an attenuator instead of a selector
            configuration.boolean(
                "USE_AIN_A",
                False
            ),
            configuration.boolean(
                "USE_AIN_B",
                False
            ),
        ]

    def main(self):
        TEXT_TOP = CHAR_HEIGHT
        BITS_LEFT = CHAR_WIDTH * 6

        N_STEPS = 256
        N_SAMPLES = 200

        while True:
            cv = ain.percent(samples=N_SAMPLES)

            if self.config.USE_AIN_A:
                atten = k1.percent(samples=N_SAMPLES)
                n1 = round(cv * atten * N_STEPS)
                if n1 == N_STEPS:
                    # prevent bounds problems since percent() returns [0, 1], not [0, 1)
                    n1 = N_STEPS - 1
            else:
                n1 = k1.read_position(steps=N_STEPS, samples=N_SAMPLES)

            if self.config.USE_AIN_B:
                atten = k2.percent(samples=N_SAMPLES)
                n2 = round(cv * atten * N_STEPS)
                if n2 == N_STEPS:
                    n2 = N_STEPS - 1
            else:
                n2 = k2.read_position(steps=N_STEPS, samples=N_SAMPLES)

            self.sequencers[0].change_sequence(n1)
            self.sequencers[1].change_sequence(n2)

            oled.fill(0)
            for i in range(len(self.sequencers)):
                s = self.sequencers[i]

                # Set the output voltages if needed
                if s.output_dirty:
                    s.apply_output()

                # Show the sequence number, sequence, and draw a box around the active bit
                oled.text(f"{s.sequence_n:5} {s.binary_sequence:08b}", 0, CHAR_HEIGHT*i + TEXT_TOP, 1)
                oled.fill_rect(
                    BITS_LEFT + s.step * CHAR_WIDTH,
                    CHAR_HEIGHT*i + TEXT_TOP,
                    CHAR_WIDTH,
                    CHAR_HEIGHT,
                    1
                )
                oled.text(
                    f"{s.current_bit}",
                    BITS_LEFT + s.step * CHAR_WIDTH,
                    CHAR_HEIGHT*i + TEXT_TOP,
                    0
                )

            oled.show()


if __name__ == "__main__":
    IttyBitty().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from europi_script import EuroPiScript

BLANK_LETTER = "."

morse_alphabet = {
    BLANK_LETTER: "",
    "A": ".-",
    "B": "-...",
    "C": "-.-.",
    "D": "-..",
    "E": ".",
    "F": "..-.",
    "G": "--.",
    "H": "....",
    "I": "..",
    "J": ".---",
    "K": "-.-",
    "L": ".-..",
    "M": "--",
    "N": "-.",
    "O": "---",
    "P": ".--.",
    "Q": "--.-",
    "R": ".-.",
    "S": "...",
    "T": "-",
    "U": "..-",
    "V": "...-",
    "W": ".--",
    "X": "-..-",
    "Y": "-.--",
    "Z": "--..",
    "0": "-----",
    "1": ".----",
    "2": "..---",
    "3": "...--",
    "4": "....-",
    "5": ".....",
    "6": "-....",
    "7": "--...",
    "8": "---..",
    "9": "----.",
}

valid_letters = [
    BLANK_LETTER,
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "J",
    "K",
    "L",
    "M",
    "N",
    "O",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
    "0",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
]

class Morse(EuroPiScript):
    dot_high_out = cv1
    dot_high_latched_out = cv2
    dot_low_out = cv4
    dot_low_latched_out = cv5
    end_of_letter_out = cv3
    end_of_word_out = cv6

    def __init__(self):
        super().__init__()

        default_letters = [BLANK_LETTER] * 16
        saved_cfg = self.load_state_json()
        self.letters = saved_cfg.get("letters", default_letters)

        # the index of the letter the user is hovering over
        self.hover_letter = 0

        self.ui_dirty = True
        self.gate_recvd = False
        self.sequence_counter = 0

        @b1.handler
        def on_b1_press():
            # advance the letter 1 place down
            idx = valid_letters.index(self.letters[self.hover_letter])
            idx = (idx - 1) % len(valid_letters)
            self.letters[self.hover_letter] = valid_letters[idx]
            self.ui_dirty = True

        @b2.handler
        def on_b2_press():
            # advance the letter 1 place up
            idx = valid_letters.index(self.letters[self.hover_letter])
            idx = (idx + 1) % len(valid_letters)
            self.letters[self.hover_letter] = valid_letters[idx]
            self.ui_dirty = True

        @din.handler
        def on_din_rise():
            self.gate_recvd = True

        @din.handler_falling
        def on_din_fall():
            # turn off the unlatched CVs
            self.dot_low_out.off()
            self.dot_high_out.off()
            self.end_of_letter_out.off()
            self.end_of_word_out.off()

    def save_state(self):
        cfg = {
            "letters": self.letters
        }
        self.save_state_json(cfg)

    def expand_sequence(self):
        """
        Expand the more representation to a boolean sequence

        We treat "." as True and "-" as False
        """
        seq = []
        for ch in self.letters:
            if ch is not None:
                for bit in morse_alphabet[ch]:
                    if bit == ".":
                        seq.append(True)
                    else:
                        seq.append(False)
        return seq

    def get_letter_ends(self):
        seq = []
        n = 0
        for ch in self.letters:
            if ch is not None:
                n += len(morse_alphabet[ch])
                seq.append(n)
        return seq

    def draw(self):
        top = OLED_HEIGHT // 2 - CHAR_HEIGHT // 2
        left = OLED_WIDTH // 2 - (CHAR_WIDTH * len(self.letters)) // 2

        oled.fill(0)

        w = ''
        for ch in self.letters:
            if ch is None:
                w = w + "."
            else:
                w = w + ch

        oled.text(
            w,
            left,
            top,
            1
        )

        oled.line(
            left + self.hover_letter * CHAR_WIDTH,
            top + CHAR_HEIGHT + 1,
            left + (self.hover_letter + 1) * CHAR_WIDTH,
            top + CHAR_HEIGHT + 1,
            1
        )

        oled.show()

    def main(self):
        while True:
            letter = int(k2.percent() * len(self.letters))
            if letter == len(self.letters):
                # .percent() can return 1, so avoid out-of-bounds issues
                letter -= 1
            if letter != self.hover_letter:
                self.hover_letter = letter
                self.ui_dirty = True

            if self.ui_dirty:
                self.ui_dirty = False
                self.draw()

            sequence = self.expand_sequence()
            end_of_letters = self.get_letter_ends()
            if len(sequence) == 0:
                turn_off_all_cvs()
                continue

            # kick out if there's no gate to process
            if not self.gate_recvd:
                continue

            self.gate_recvd = False

            # if we've modified the sequence, check for overflow
            if self.sequence_counter >= len(sequence):
                self.sequence_counter = 0

            # play the sequence
            if sequence[self.sequence_counter]:
                self.dot_high_out.on()
                self.dot_high_latched_out.on()
                self.dot_low_out.off()
                self.dot_low_latched_out.off()
            else:
                self.dot_high_out.off()
                self.dot_high_latched_out.off()
                self.dot_low_out.on()
                self.dot_low_latched_out.on()

            # increment
            self.sequence_counter += 1

            # check for end of letter
            if self.sequence_counter in end_of_letters:
                self.end_of_letter_out.on()

            # check for end of word
            if self.sequence_counter == len(sequence):
                self.end_of_word_out.on()


if __name__ == "__main__":
    Morse().main()



#!/usr/bin/env python3
# Copyright 2023 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implements Conway's Game of Life as a pseudo-random LFO kernel

Outputs 1-3 are 0-10V control voltages, outputs 4-6 are 5V gates

@author Chris I-B <ve4cib@gmail.com>
@year   2023

Throughout the program we frequently use `>> 3` instead of `//8` for doing integer division when converting
from bit indices to byte indices.
"""

from europi import *
from europi_script import EuroPiScript
from experimental.bitarray import *
from random import random as rnd

import math

# We re-use this constant a lot, so just save it for easy re-use
LOG2 = math.log(2)

# How many pixels are on the screen
NUM_PIXELS = OLED_HEIGHT * OLED_WIDTH


def stdev(l):
    """Return the standard deviation of a list of values

    @param l  The list of numbers we want to calculate the standard deviation of

    @return The standard deviation of the values in @l
    """
    mean = sum(l)/len(l)
    return ( sum([((x - mean) ** 2) for x in l]) / len(l) )**0.5


def bitwise_entropy(arr):
    """Calculate the entropy of the bit string in a bytearray

    @param arr  A bytearray, treated as a bit string

    @return the Shannon Entropy of the string, assuming a 50/50 chance of any bit being 1 or 0
    """
    # Count how many bits are 1 in the whole bytearray
    count1s = 0
    for b in arr:
        for i in range(8):
            if b & (1 << i):
                count1s += 1

    # Make sure we don't have all-1 or all-0 in the array; handle those cases
    num_bits = len(arr) << 3
    if count1s == 0:
        return 0.0
    elif count1s == num_bits:
        return 1.0
    else:
        # Calculate the entropy of the string
        # E = sum(p(x) * log_2(p(x))) = sum(p(x) * log(p(x))) / log(2)
        prob_1 = count1s / num_bits
        p_x = [
            1.0 - prob_1,
            prob_1
        ]
        return -sum([ p * math.log(p) for p in p_x]) / LOG2


class Conway(EuroPiScript):
    def __init__(self):
        # For ease of blitting, store the field as a bit array
        # Each byte is 8 horizontally adjacent pixels, with the most significant bit
        # on the left
        self.field = bytearray(NUM_PIXELS >> 3)
        self.next_field = bytearray(NUM_PIXELS >> 3)

        # Keep 2 separate frame buffer instances so we don't need to recreate the FB objects when we draw
        self.frame = FrameBuffer(self.field, OLED_WIDTH, OLED_HEIGHT, MONO_HLSB)
        self.next_frame = FrameBuffer(self.next_field, OLED_WIDTH, OLED_HEIGHT, MONO_HLSB)

        # Simple optimization; keep a list of spaces whose states changed & their neighbours
        # This is initially empty as the field is entirely blank
        # to save memory instead of using a List or Set object, we use a bit array with a length equal to
        # the size of the field
        self.num_changes = 0
        self.changed_spaces = bytearray(NUM_PIXELS >> 3)
        self.next_changed_spaces = bytearray(NUM_PIXELS >> 3)

        # how many cells were born this tick?
        self.num_born = 0

        # how many cells died this tick?
        self.num_died = 0

        # how many cells are currently alive?
        # this gets updated on every tick and on random spawns
        self.num_alive = 0

        # Set to True if we want to clear the field & respawn
        self.reset_requested = False

        # statically allocated array of the indices of the cells around the one we're considering
        # this is just an optimization to avoid re-allocating this array
        self.neighbourhood = [0, 0, 0, 0, 0, 0, 0, 0]

        # keep the last few changes in population in a list to check if it's oscillating predictably
        self.population_deltas = []
        self.MAX_DELTAS = 12

        # statically allocated array we use to store the sums of cells when checking for statis
        self.statis_sums = [0] * self.MAX_DELTAS

        @b1.handler
        def on_b1():
            self.reset_requested = True

        @b2.handler
        def on_b2():
            self.reset_requested = True

        @din.handler
        def on_din():
            self.reset_requested = True

    def get_neigbour_indices(self, index):
        """Get the indices of the 8 bits adjacent to the given index

        If we're on the top/left/bottom/right edge, wrap arround to the opposite row/column, treating the world
        as a torus

        Unfortunately we don't have enough RAM to save this in a lookup table, so we have to recalculate it
        every time, which slows down the simulation
        """
        row = index // OLED_WIDTH
        col = index % OLED_WIDTH

        def rowcol2index(r, c):
            return (r % OLED_HEIGHT) * OLED_WIDTH + (c % OLED_WIDTH)

        n = 0
        for i in range(-1, 2):
            for j in range(-1, 2):
                if i != 0 or j != 0:
                    self.neighbourhood[n] = rowcol2index(row+i, col+j)
                    n += 1

    def calculate_spawn_level(self):
        """Calculate what percentage of the field should contain new cells

        We want to avoid having the spawn rate _too_ high as that can result in high CPU loads and RAM usage.
        - K1: [0, 0.5]
        - K2: [-0.5, 0.5]
        - AIN: [0, 1]

        P = K1 + K2 * AIN => [-0.5, 1.0]
        """
        base_spawn_level = k1.percent() / 2

        # get the level of AIN, attenuverted by K2
        cv_mod = ain.percent()/2
        cv_att = k2.percent() - 1
        cv_mod = cv_mod * cv_att

        spawn_level = clamp(base_spawn_level + cv_mod, 0, 1)
        return spawn_level

    def reset(self):
        """Clear the whole field and spawn random data in it
        """
        for i in range(len(self.field)):
            self.next_field[i] = 0x00

        self.num_alive = 0
        self.population_deltas = []

        # fill the field with random cells
        fill_level = self.calculate_spawn_level()
        for i in range(NUM_PIXELS):
            x = rnd()
            is_alive = get_bit(self.field, i)
            if x < fill_level and not is_alive:
                # if the space isn't already filled and we want to fill it
                set_bit(self.field, i, True)
                set_bit(self.next_field, i, True)
                self.num_alive += 1
            elif x >= fill_level and is_alive:
                # if the space is filled and we want to clear it
                set_bit(self.field, i, False)
                set_bit(self.next_field, i, False)
                self.num_alive -= 1

        # Assume the whole field has changed
        set_all_bits(self.changed_spaces, True)
        self.num_changes = NUM_PIXELS

    def draw(self):
        """Show the current playing field on the OLED
        """
        oled.blit(self.frame, 0, 0)
        oled.show()

    def tick(self):
        """Calculate the state of the next generation

        This checks the regions around every changed space in the previous generation, updating the
        total population and counting how many births & deaths this generation had.

        If a reset was requested, the field is cleared & randomly reset _before_ calculating the new generation
        """
        self.num_born = 0
        self.num_died = 0
        self.num_changes = 0

        if self.reset_requested:
            self.reset_requested = False
            self.reset()

        for bit_index in range(NUM_PIXELS):
            if not get_bit(self.changed_spaces, bit_index):
                continue

            self.get_neigbour_indices(bit_index)
            num_neighbours = sum(1 for n in self.neighbourhood if get_bit(self.field, n))

            if get_bit(self.field, bit_index):
                if num_neighbours == 2 or num_neighbours == 3:        # happy cell, stays alive
                    set_bit(self.next_field, bit_index, True)
                else:                                                 # sad cell, dies
                    set_bit(self.next_field, bit_index, False)
                    self.num_died += 1
                    self.num_alive -= 1

                    self.num_changes += 1
                    set_bit(self.next_changed_spaces, bit_index, 1)
                    for n in self.neighbourhood:
                        set_bit(self.next_changed_spaces, n, 1)
            else:
                if num_neighbours == 3:                               # baby cell is born!
                    set_bit(self.next_field, bit_index, True)
                    self.num_alive += 1
                    self.num_born += 1

                    self.num_changes += 1
                    set_bit(self.next_changed_spaces, bit_index, 1)
                    for n in self.neighbourhood:
                        set_bit(self.next_changed_spaces, n, 1)
                else:                                                 # empty space remains empty
                    set_bit(self.next_field, bit_index, False)

        # swap field & next_field so we don't need to copy between arrays
        tmp = self.next_field
        self.next_field = self.field
        self.field = tmp

        tmp = self.next_frame
        self.next_frame = self.frame
        self.frame = tmp

        tmp = self.next_changed_spaces
        self.next_changed_spaces = self.changed_spaces
        self.changed_spaces = tmp
        set_all_bits(self.next_changed_spaces)

    def check_for_stasis(self):
        """Check the population changes over time to see if we've reached a state of stasis
        """
        # we must have at least MAX_DELTAS generations of data
        if len(self.population_deltas) < self.MAX_DELTAS:
            return False

        # if there are no changes or everything is dead, we've reached static stasis
        if self.num_changes == 0 or self.num_alive == 0:
            return True

        # if the population is oscillating up and down predicatbly, we've probably reached stasis
        # check for 2, 3, and 4 step repetitions
        for pattern_length in range(2, 5):
            count = self.MAX_DELTAS // pattern_length
            for i in range(count):
                self.statis_sums[i] = sum(self.population_deltas[i*pattern_length:i*pattern_length+pattern_length])

            # check the standard deviation
            deviation = stdev(self.statis_sums[0:count])
            mean = sum(self.statis_sums[0:count])/count
            if deviation <= 1 and abs(mean) <= 1:
                return True

        return False

    def main(self):
        """The main loop for the program

        Handles setting the CV output, drawing to the OLED, and triggering the simulation
        """
        turn_off_all_cvs()
        self.reset()

        in_stasis = False

        while True:
            # turn off the stasis gate while we calculate the next generation
            cv6.off()

            # turn on the FPS gate when we start calculating
            cv4.on()

            # calculate the next generation
            self.tick()

            # turn off the FPS gate when we're done calculating but before we draw
            cv4.off()

            # show the results on the OLED
            self.draw()

            # check for stasis conditions
            self.population_deltas.append(self.num_born - self.num_died)
            if len(self.population_deltas) > self.MAX_DELTAS:
                self.population_deltas.pop(0)
            in_stasis = self.check_for_stasis()

            cv1.voltage(MAX_OUTPUT_VOLTAGE * bitwise_entropy(self.field))
            if self.num_born > self.num_died:
                cv5.on()
            else:
                cv5.off()

            # Make sure we don't divide by zero
            if self.num_alive > 0:
                cv2.voltage(MAX_OUTPUT_VOLTAGE * self.num_born / self.num_alive)
            else:
                cv2.off()


            # Prevent values greater than 1 & division-by-zero errors
            hi = max(self.num_died, self.num_born)
            low = min(self.num_died, self.num_born)
            if (hi > 0):
                cv3.voltage(MAX_OUTPUT_VOLTAGE * (low/hi))
            else:
                cv3.off()

            # If we've achieved statis, set CV6 & trigger a reset
            if in_stasis:
                cv6.on()
                self.reset_requested = True

if __name__ == "__main__":
    Conway().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from machine import ADC
from time import sleep

from europi import k1, k2, b1, b2, oled, MAX_UINT16
from europi_script import EuroPiScript
from experimental.knobs import KnobBank

"""
An example program showing the use of a KnobBank or LockableKnobs. This script is not meant to be merged into main,
it only exists so that PR reviewers can try out a LockableKnob in physical hardware easily.
"""


class KnobPlayground(EuroPiScript):
    def __init__(self):
        super().__init__()
        self.next_k1 = False
        self.next_k2 = False

        self.kb1 = (
            KnobBank.builder(k1)
            .with_locked_knob("p1", initial_uint16_value=0, threshold_percentage=0.02)
            .with_locked_knob("p2", initial_uint16_value=MAX_UINT16 / 5)
            .with_locked_knob("p3", initial_uint16_value=MAX_UINT16 / 3)
            .build()
        )
        self.kb2 = (
            KnobBank.builder(k2)
            .with_disabled_knob()
            .with_locked_knob("p4", initial_percentage_value=0.5, threshold_from_choice_count=7)
            .with_locked_knob("p5", initial_percentage_value=1, threshold_from_choice_count=3)
            .build()
        )

        @b1.handler
        def next_knob1():
            self.next_k1 = True

        @b2.handler
        def next_knob2():
            self.next_k2 = True

    def main(self):
        choice_p4 = ["a", "b", "c", "d", "e", "f", "g"]
        choice_p5 = ["one", "two", "three"]

        while True:
            if self.next_k1:
                self.kb1.next()
                self.next_k1 = False
            if self.next_k2:
                self.kb2.next()
                self.next_k2 = False

            p1 = "X" if self.kb1.index == 0 else " "
            p2 = "*" if self.kb1.index == 1 else " "
            p3 = "*" if self.kb1.index == 2 else " "
            pd = "*" if self.kb2.index == 0 else " "
            p4 = "*" if self.kb2.index == 1 else " "
            p5 = "*" if self.kb2.index == 2 else " "
            text = (
                f"{p1} {self.kb1.p1.range(1000):4}  {pd} {k2.range()}      \n"
                + f"{p2} {int(round(self.kb1.p2.percent(), 2)*100):3}%  {p4} {self.kb2.p4.choice(choice_p4):5}\n"
                + f"{p3} {self.kb1.p3.read_position():4}  {p5} {self.kb2.p5.choice(choice_p5):5}"
            )
            oled.centre_text(text)


if __name__ == "__main__":
    KnobPlayground().main()



# Copyright 2023 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logic gates for the EuroPi

Treats ain as a digital input with a threshold of 0.8V

@author Chris Iverach-Brereton <ve4cib@gmail.com>
@date   2023-02-13
"""

from europi import *
from europi_script import EuroPiScript

from experimental.a_to_d import AnalogReaderDigitalWrapper
from experimental.screensaver import OledWithScreensaver

ssoled = OledWithScreensaver()

class Logic(EuroPiScript):
    """The main workhorse of the whole module

    Does all 6 basic logic operations, sending the result to each of the
    6 outputs
    """
    def __init__(self):
        super().__init__()

        # the six logic operations' outputs, as 0/1 values
        # this lets us use bitwise math operations to set the output voltages to either 0 or 1V
        # as well as satisfying the input requirements for the europi.Output.value function
        self.x_and_y = 0
        self.x_or_y = 0
        self.x_xor_y = 0
        self.x_nand_y = 0
        self.x_nor_y = 0
        self.x_xnor_y = 0

        self.din1 = din
        self.din2 = AnalogReaderDigitalWrapper(ain)

        # connect ISRs
        self.din1.handler(self.update_logic)
        self.din1.handler_falling(self.update_logic)
        self.din2.handler(self.update_logic)
        self.din2.handler_falling(self.update_logic)
        b1.handler(self.button_state_change)
        b1.handler_falling(self.button_state_change)
        b2.handler(self.button_state_change)
        b2.handler_falling(self.button_state_change)

    def button_state_change(self):
        self.update_logic()
        ssoled.notify_user_interaction()

    @classmethod
    def display_name(cls):
        return "Logic"

    def update_logic(self):
        """Updates the values of our 6 logic operations
        """
        x = self.din1.value() | b1.value()
        y = self.din2.value() | b2.value()

        self.x_and_y = x & y
        self.x_or_y = x | y
        self.x_xor_y = x ^ y
        self.x_nand_y = abs(self.x_and_y - 1) # bit of a hack to get the inverted values
        self.x_nor_y = abs(self.x_or_y -1)    # ~ results in some -2 results, which we don't want
        self.x_xnor_y = abs(self.x_xor_y -1)  # so some simple int math will suffice

    def main(self):
        knob_wakeup_threshold = 0.05

        prev_k1 = k1.percent()
        prev_k2 = k2.percent()

        while True:
            # update ain
            self.din2.update()

            # check to see if we've wiggled a knob to exit the screensaver
            current_k1 = k1.percent()
            current_k2 = k2.percent()

            if (
                abs(current_k1 - prev_k1) >= knob_wakeup_threshold or
                abs(current_k2 - prev_k2) >= knob_wakeup_threshold
            ):
                ssoled.notify_user_interaction()
                prev_k1 = current_k1
                prev_k2 = current_k2

            # set the output voltages
            cv1.value(self.x_and_y)
            cv2.value(self.x_or_y)
            cv3.value(self.x_xor_y)
            cv4.value(self.x_nand_y)
            cv5.value(self.x_nor_y)
            cv6.value(self.x_xnor_y)

            display_txt = " &:{0}  |:{1}  ^:{2}\n!&:{3} !|:{4} !^:{5}".format(
                self.x_and_y,
                self.x_or_y,
                self.x_xor_y,
                self.x_nand_y,
                self.x_nor_y,
                self.x_xnor_y
            )

            ssoled.centre_text(display_txt)

if __name__ == "__main__":
    Logic().main()



# Copyright 2023 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Equal-temperment quantizer for the EuroPi

Features configurable intervals for multiple outputs and customizable scale

  @author Chris Iverach-Brereton <ve4cib@gmail.com>
  @date   2023-02-12
"""

from europi import *
from europi_script import EuroPiScript

from experimental.quantizer import Quantizer, VOLTS_PER_OCTAVE, VOLTS_PER_SEMITONE, SEMITONES_PER_OCTAVE
from experimental.screensaver import Screensaver

import time

## Whe in triggered mode we only quantize when we receive an external clock signal
MODE_TRIGGERED=0

## In continuous mode the digital input is ignored and we quantize the input
#  at the highest rate possible
MODE_CONTINUOUS=1

## How many milliseconds of idleness do we need before we trigger the screensaver?
#
#  =20 minutes
SCREENSAVER_TIMEOUT_MS = 1000 * 60 * 20


SELECT_OPTION_Y = 16
HALF_CHAR_WIDTH = int(CHAR_WIDTH / 2)

class ScreensaverScreen(Screensaver):
    """Blank the screen when idle

    Eventually it might be neat to have an animation, but that's
    not necessary for now
    """
    def __init__(self, quantizer):
        super().__init__()
        self.quantizer = quantizer

    def on_button1(self):
        self.quantizer.active_screen = self.quantizer.kb

class KeyboardScreen:
    """Draws a pretty keyboard and indicates what notes are enabled
    and what note is being played as the primary output
    """

    def __init__(self, quantizer):
        self.quantizer = quantizer
        self.highlight_note = 0

        # X, Y, bw
        self.enable_marks = [
            (  8, 2, 0),
            ( 17, 2, 1),
            ( 26, 2, 0),
            ( 35, 2, 1),
            ( 43, 2, 0),
            ( 59, 2, 0),
            ( 69, 2, 1),
            ( 77, 2, 0),
            ( 85, 2, 1),
            ( 94, 2, 0),
            (103, 2, 1),
            (112, 2, 0)
        ]

        self.playing_marks = [
            (  8, 20, 0),
            ( 17, 15, 1),
            ( 26, 20, 0),
            ( 35, 15, 1),
            ( 43, 20, 0),
            ( 59, 20, 0),
            ( 69, 15, 1),
            ( 77, 20, 0),
            ( 85, 15, 1),
            ( 94, 20, 0),
            (103, 15, 1),
            (112, 20, 0)
        ]

    def draw(self):
        # a 128x32 keyboard image
        # see https://github.com/Allen-Synthesis/EuroPi/blob/main/software/oled_tips.md
        # and https://github.com/novaspirit/img2bytearray
        KB_HEIGHT = 32
        KB_WIDTH = 128
        kb=b'\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xc0\x7f\xe0?\xfe\xff\xf8\x0f\xfc\x07\xfe\x03\xff\xe0\x07\xff\xfb\xff\xfd\xff\xfe\xff\xff\x7f\xff\xbf\xff\xdf\xff\xe0\x07\xff\xfb\xff\xfd\xff\xfe\xff\xff\x7f\xff\xbf\xff\xdf\xff\xe0\x07\xff\xfb\xff\xfd\xff\xfe\xff\xff\x7f\xff\xbf\xff\xdf\xff\xe0\x07\xff\xfb\xff\xfd\xff\xfe\xff\xff\x7f\xff\xbf\xff\xdf\xff\xe0\x07\xff\xfb\xff\xfd\xff\xfe\xff\xff\x7f\xff\xbf\xff\xdf\xff\xe0\x07\xff\xfb\xff\xfd\xff\xfe\xff\xff\x7f\xff\xbf\xff\xdf\xff\xe0\x07\xff\xfb\xff\xfd\xff\xfe\xff\xff\x7f\xff\xbf\xff\xdf\xff\xe0\x07\xff\xfb\xff\xfd\xff\xfe\xff\xff\x7f\xff\xbf\xff\xdf\xff\xe0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'

        # read the encoder value from knob 1 so we know what key to highlight
        self.highlight_note = k1.range(len(self.quantizer.scale))

        # draw the keyboard image to the screen
        img = bytearray(kb)
        imgFB = FrameBuffer(img, KB_WIDTH, KB_HEIGHT, MONO_HLSB)
        oled.blit(imgFB,0,0)

        # mark the enabled notes with a .
        for i in range(len(self.quantizer.scale)):
            if self.quantizer.scale[i]:
                oled.text('.', *self.enable_marks[i])

        # mark the active note with a +
        k = self.quantizer.current_note
        oled.text('+', self.playing_marks[k][0], self.playing_marks[k][1], self.playing_marks[k][2])

        # clear the bottom of the screen and mark the togglable key with a line
        oled.fill_rect(0, OLED_HEIGHT-2, OLED_WIDTH, 2, 0)
        oled.fill_rect(self.enable_marks[self.highlight_note][0], OLED_HEIGHT-1, 7, 1, 1)

        oled.show()

    def on_button1(self):
        self.quantizer.scale[self.highlight_note] = not self.quantizer.scale[self.highlight_note]
        self.quantizer.save()

class MenuScreen:
    """Advanced menu options screen
    """

    def __init__(self, quantizer):
        self.quantizer = quantizer
        self.menu_item = 0

        self.menu_items = [
            ModeChooser(quantizer),
            RootChooser(quantizer),
            OctaveChooser(quantizer),
            IntervalChooser(quantizer, 2),
            IntervalChooser(quantizer, 3),
            IntervalChooser(quantizer, 4),
            IntervalChooser(quantizer, 5)
        ]

    def draw(self):
        self.menu_item = k1.range(len(self.menu_items))
        self.menu_items[self.menu_item].draw()

    def on_button1(self):
        self.menu_items[self.menu_item].on_button1()

class ModeChooser:
    """Used by MenuScreen to choose the operating mode
    """
    def __init__(self, quantizer):
        self.quantizer = quantizer

        self.mode_names = [
            "Triggered",
            "Continuous"
        ]

    def read_mode(self, mode='integer'):
        if mode == 'string':
            return k2.choice(self.mode_names)
        else:
            return k2.range(len(self.mode_names))

    def on_button1(self):
        new_mode = self.read_mode()
        self.quantizer.mode = new_mode
        self.quantizer.save()

    def draw(self):
        oled.fill(0)
        oled.text(f"Mode", 0, 0)

        current_mode = self.mode_names[self.quantizer.mode]
        new_mode = self.read_mode(mode='string')
        QuantizerScript.choose_option(self, new_mode, current_mode, self.mode_names)

        oled.show()

class RootChooser:
    """Used by MenuScreen to choose the transposition offset
    """
    def __init__(self, quantizer):
        self.quantizer = quantizer
        self.root_names = [
            "C ",
            "C#",
            "D ",
            "D#",
            "E ",
            "F ",
            "F#",
            "G ",
            "G#",
            "A ",
            "A#",
            "B "
        ]

    def read_root(self, mode='integer'):
        if mode == 'string':
            return k2.choice(self.root_names)
        else:
            return k2.range(len(self.root_names))

    def on_button1(self):
        new_root = self.read_root()
        self.quantizer.root = new_root
        self.quantizer.save()

    def draw(self):
        oled.fill(0)
        oled.text(f"Transpose", 0, 0)

        new_root = self.read_root(mode='string')
        current_root = self.root_names[self.quantizer.root]
        QuantizerScript.choose_option(self, new_root, current_root, self.root_names)

        oled.show()

class OctaveChooser:
    """Used by MenuScreen to choose the octave offset
    """
    def __init__(self, quantizer):
        self.quantizer = quantizer
        self.octave_texts = ['-4', '-3', '-2', '-1', '0', '+1', '+2', '+3', '+4']
        self.octave_text_y = 12

    def read_octave(self, mode='integer'):
        if mode == 'string':
            return k2.choice(self.octave_texts)
        else:
            return k2.range(9) - 4  # result should be -1 to +2

    def on_button1(self):
        new_octave = self.read_octave()
        self.quantizer.octave = new_octave
        self.quantizer.save()

    def draw(self):
        oled.fill(0)
        oled.text(f"Octave", 0, 0)

        new_octave = self.read_octave(mode='string')
        current_octave = self.quantizer.octave

        if current_octave > 0:
            current_octave = '+' + str(current_octave)
        else:
            current_octave = str(current_octave)

        QuantizerScript.choose_option(self, new_octave, current_octave, self.octave_texts)

        oled.show()

class IntervalChooser:
    """Used by MenuScreen to choose the interval offset for a given output
    """
    def __init__(self, quantizer, n):
        self.quantizer = quantizer
        self.n = n
        self.interval_names = [
            "-Pe 8",
            "-MA 7",
            "-mi 7",
            "-MA 6",
            "-mi 6",
            "-Pe 5",
            "-Di 5",
            "-Pe 4",
            "-MA 3",
            "-mi 3",
            "-MA 2",
            "-mi 2",
            "Pe 1",
            "+mi 2",
            "+MA 2",
            "+mi 3",
            "+MA 3",
            "+Pe 4",
            "+Di 5",
            "+Pe 5",
            "+mi 6",
            "+MA 6",
            "+mi 7",
            "+MA 7",
            "+Pe 8"
        ]


    def read_interval(self, mode='integer'):
        if mode == 'string':
            return k2.choice(self.interval_names)
        else:
            return k2.range(len(self.interval_names)) - 12

    def on_button1(self):
        new_interval = self.read_interval()
        self.quantizer.intervals[self.n-2] = new_interval
        self.quantizer.save()

    def draw(self):
        oled.fill(0)
        oled.text(f"Output {self.n}", 0, 0)

        new_interval = self.read_interval(mode='string')
        current_interval = self.interval_names[self.quantizer.intervals[self.n-2]+12]
        QuantizerScript.choose_option(self, new_interval, current_interval, self.interval_names)

        oled.show()

class QuantizerScript(EuroPiScript):
    """The main EuroPi program. Uses Scale to quantize incoming analog voltages
    and round them to the nearest note on the scale.

    Primary output is on cv1, with cv2-5 as aux outputs shifted up/down a fixed
    number of semitones.  cv6 outputs a gate/trigger.
    """
    def __init__(self):
        super().__init__()

        # keep track of the last time the user interacted with the module
        # if we're idle for too long, start the screensaver
        self.last_interaction_time = time.ticks_ms()

        # Continious quantizing, or only on an external trigger?
        self.mode = MODE_TRIGGERED

        # What semitone is the root of the scale?
        # 0 = C, 1 = C#/Db, 2 = D, etc...
        # This is used to transpose the output up the given number of semitones
        self.root = 0

        # What octave are we outputting?
        self.octave = 0

        # Outputs 2-5 output the same note, shifted up or down by
        # a fixed number of semitones
        self.intervals = [0, 0, 0, 0]
        self.aux_outs = [cv2, cv3, cv4, cv5]

        # The current scale we're quantizing to
        self.scale = Quantizer()

        # The input/output voltages
        self.input_voltage = 0.0
        self.output_voltage = 0.0

        # The semitone we're currently outputting on cv1 (0-11)
        self.current_note = 0

        # GUI/user interaction
        self.kb = KeyboardScreen(self)
        self.menu = MenuScreen(self)
        self.screensaver = ScreensaverScreen(self)
        self.active_screen = self.kb

        self.screen_centre = int(OLED_WIDTH / 2)

        self.load()

        # connect event handlers for the rising & falling clock edges + button presses

        @din.handler
        def on_rising_clock():
            """Handler for the rising edge of the input clock
            """
            if self.mode == MODE_TRIGGERED:
                self.read_quantize_output()
                cv6.on()

        @din.handler_falling
        def on_falling_clock():
            """Handler for the falling edge of the input clock
            """
            if self.mode == MODE_TRIGGERED:
                cv6.off()

        @b1.handler
        def on_b1_press():
            """Handler for pressing button 1

            Button 1 is used for the main interaction and is passed to
            the current display for user interaction
            """
            self.last_interaction_time = time.ticks_ms()
            self.active_screen.on_button1()

        @b2.handler
        def on_b2_press():
            """Handler for pressing button 2

            Button 2 is used to cycle between screens
            """
            self.last_interaction_time = time.ticks_ms()

            if self.active_screen == self.kb:
                self.active_screen = self.menu
            else:
                self.active_screen = self.kb

    def load(self):
        """Load the persistent settings from storage and apply them
        """
        state = self.load_state_json()

        loaded_scale = state.get("scale", [True]*12)  # default to a chromatic scale
        self.scale.notes = loaded_scale

        self.root = state.get("root", self.root)
        self.octave = state.get("octave", self.octave)
        self.intervals = state.get("intervals", self.intervals)
        self.mode = state.get("mode", self.mode)

    def save(self):
        """Save the current settings to persistent storage
        """
        state = {
            "scale": self.scale.notes,
            "root": self.root,
            "octave": self.octave,
            "intervals": self.intervals,
            "mode": self.mode
        }
        self.save_state_json(state)

    @classmethod
    def display_name(cls):
        return "Quantizer"

    def quantize(self, analog_in):
        """Take an analog signal and process it

        Sets self.current_note and self.output_voltage
        """

        (volts, semitone) = self.scale.quantize(analog_in)

        # apply our octave & transposition offsets
        volts = volts + self.octave * VOLTS_PER_OCTAVE + self.root * VOLTS_PER_SEMITONE

        self.output_voltage = volts
        self.current_note = semitone

    def read_quantize_output(self):
        """Read the input signal, quantize it, set outputs 1-5 accordingly

        Called by the main loop in continuous mode or the rising clock handler
        in triggered mode
        """
        self.input_voltage = ain.read_voltage(500)   # increase the number of samples to help reduce noise
        self.quantize(self.input_voltage)

        cv1.voltage(self.output_voltage)

        for i in range(len(self.aux_outs)):
            self.aux_outs[i].voltage(self.output_voltage + self.intervals[i] * VOLTS_PER_SEMITONE)

    def choose_option(self, new_item, current_item, all_items):
        item_widths = []
        for item_text in all_items:
            if item_text == new_item:
                offset = -int(sum(item_widths) + (CHAR_WIDTH * (len(item_widths))) + (CHAR_WIDTH * (len(item_text) / 2)) - (OLED_WIDTH / 2))
            item_widths.append(len(item_text) * CHAR_WIDTH)

        x = offset

        for index, item_text in enumerate(all_items):
            item_text_width = item_widths[index]
            if item_text == current_item:
                oled.fill_rect((x - 1), SELECT_OPTION_Y, (item_text_width + 3), (CHAR_HEIGHT + 4), 1)
                oled.text(item_text, x, (SELECT_OPTION_Y + 2), 0)
            elif item_text == new_item:
                oled.rect((x - 1), SELECT_OPTION_Y, (item_text_width + 3), (CHAR_HEIGHT + 4), 1)
                oled.text(item_text, x, (SELECT_OPTION_Y + 2), 1)
            else:
                oled.text(item_text, x, (SELECT_OPTION_Y + 2), 1)
            x += item_text_width + CHAR_WIDTH

    def main(self):
        """The main loop; reads from ain, sets the output voltages
        """

        while True:
            # Check if we've been idle for too long; if so, blank the screen
            # to prevent burn-in
            now = time.ticks_ms()
            if time.ticks_diff(now, self.last_interaction_time) > SCREENSAVER_TIMEOUT_MS:
                self.active_screen = self.screensaver

            if self.mode == MODE_CONTINUOUS:
                # clear the previous trigger
                cv6.off()

                # Read the new voltage and output it
                last_output = self.output_voltage
                self.read_quantize_output()

                if last_output != self.output_voltage:
                    cv6.on()

                # In continuous mode we fall back to a 100Hz internal clock
                # that effectively simulates a very high-speed input trigger
                # Note that this rate has to be long enough for the trigger on
                # cv6 to be useful to other modules.  A 10ms trigger is a
                # reasonable compromise between high-speed input signal processing
                # and keeping the code simple
                CYCLE_RATE = 0.01
                time.sleep(CYCLE_RATE)

            self.active_screen.draw()

if __name__ == "__main__":
    QuantizerScript().main()



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Generates random voltages based on Bezier curves

Inspired by Addac Systems' & Monotrail's ADDAC507 collaboration
"""

from europi import *
from europi_script import EuroPiScript

from experimental.knobs import *
from experimental.math_extras import solve_linear_system
from experimental.screensaver import OledWithScreensaver

import configuration
import math
import random
import time

ssoled = OledWithScreensaver()

class LogicalOperator:
    """Abstract class used for saving the gate outputs' logical operation

    Subclasses must implement the compare(a, b) method
    """
    def compare(self, a, b):
        raise Exception("Not implemented")

class LogicalAnd(LogicalOperator):
    """Binary AND operator"""
    def compare(self, a, b):
        return a and b

class LogicalOr(LogicalOperator):
    """Binary OR operator"""
    def compare(self, a, b):
        return a or b

class LogicalXor(LogicalOperator):
    """Binary XOR operator"""
    def compare(self, a, b):
        return (a or b) and not (a and b)

class LogicalNand(LogicalOperator):
    """Binary NAND operator"""
    def compare(self, a, b):
        return not (a and b)

class LogicalNor(LogicalOperator):
    """Binary Nor operator"""
    def compare(self, a, b):
        return not (a or b)

class LogicalXnor(LogicalOperator):
    """Binary XNOR operator"""
    def compare(self, a, b):
        return (a and not b) or (not a and b)


CLIP_MODE_LIMIT = 0
CLIP_MODE_FOLD = 1
CLIP_MODE_THRU = 2
N_CLIP_MODES = 3

CLIP_MODE_NAMES = [
    "Limit",
    "Fold",
    "Thru"
]

AIN_MODE_FREQUENCY = "frequency"
AIN_MODE_CURVE = "curve"


class Point2D:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __str__(self):
        return f"{self.x},{self.y}"


def linear_interpolate(x1, x2, t):
    """Linearly interpolate from one value to another

    @param x1 The initial value of x
    @param x2 The final value of x
    @param t  The time [0, 1]

    @return   The interpolated value of x
    """
    return x1 * (1-t) + x2 * t


class BezierCurve:
    """Calculates 2D bezier curves using quadratic interpolation

    The intermediate point, P1, always lies on the inverse diagonal o
    """
    def __init__(self):
        self.origin = Point2D(0, 0)
        self.next_point = Point2D(1, 0)

    def set_next_value(self, y):
        """Set the y value for the next time increment (0-1)
        """
        self.origin.y = self.next_point.y
        self.next_point.y = y

    def value_at(self, t, k):
        """Get the value of the bezier curve for a given time

        @param t  The time [0, 1]
        @param k  The curve constant, in the range [-1, 1]  See @interpolate for details on the curve constant
        """
        # Get 4 points on the curve so we can create a cubic equation for the curve
        p1 = self.interpolate(0, k)
        p2 = self.interpolate(1/3, k)
        p3 = self.interpolate(2/3, k)
        p4 = self.interpolate(1, k)

        # matrix representation
        m = [
            [p1.x**3, p1.x**2, p1.x, 1, p1.y],
            [p2.x**3, p2.x**2, p2.x, 1, p2.y],
            [p3.x**3, p3.x**2, p3.x, 1, p3.y],
            [p4.x**3, p4.x**2, p4.x, 1, p4.y],
        ]

        coeffs = solve_linear_system(m)
        return coeffs[0] * t**3 + coeffs[1] * t**2 + coeffs[2] * t + coeffs[3]


    def interpolate(self, t, k):
        """Calculate the 2D position for the given time

        See https://en.wikipedia.org/wiki/Bezier_curve#Cubic_curves for details on the math.

        @param t  The current interpolated time, between 0 and 1
        @param k  A constant indicating our progression from horizontal endpoints (smoothest) @ -1.0 to
                  linear @ 0.0 to vertical endpoints (spikiest) @ 1.0
        @return   The interpolated 2D point on the bezier curve
        """
        # Define 2 intermediate points, which are either horizontally or vertically aligned with the start and end points
        p0 = self.origin
        p1 = Point2D(0,0)
        p2 = Point2D(0,0)
        p3 = self.next_point

        if k <= 0:
            # start/endpoints aligned horizontally with intermediate points
            # /3 results in slightly smoother results; /2 was giving cusps on sharp peaks & valleys
            p1.x = p0.x - k/3
            p2.x = p3.x + k/3
            p1.y = p0.y
            p2.y = p3.y
        else:
            # start/endpoints aligned vertically with intermediate points
            p1.x = p0.x
            p2.x = p3.x

            dy = abs(p0.y - p3.y)

            if p0.y < p3.y:
                # p1 goes up, p2 goes down
                p1.y = p1.y + dy * k/2
                p2.y = p3.y - dy * k/2
            else:
                # p1 goes down, p2 goes up
                p1.y = p1.y - dy * k/2
                p2.y = p3.y + dy * k/2

        q0 = Point2D(
            linear_interpolate(p0.x, p1.x, t),
            linear_interpolate(p0.y, p1.y, t)
        )
        q1 = Point2D(
            linear_interpolate(p1.x, p2.x, t),
            linear_interpolate(p1.y, p2.y, t)
        )
        q2 = Point2D(
            linear_interpolate(p2.x, p3.x, t),
            linear_interpolate(p2.y, p3.y, t)
        )

        r0 = Point2D(
            linear_interpolate(q0.x, q1.x, t),
            linear_interpolate(q0.y, q1.y, t)
        )
        r1 = Point2D(
            linear_interpolate(q1.x, q2.x, t),
            linear_interpolate(q1.y, q2.y, t)
        )

        b = Point2D(
            linear_interpolate(r0.x, r1.x, t),
            linear_interpolate(r0.y, r1.y, t)
        )

        return b


class OutputChannel:
    """Wrapper for a CV output channel
    """
    def __init__(self, script, frequency_in, curve_in, cv_out):
        self.script = script
        self.curve = BezierCurve()

        self.cv_out = cv_out
        self.frequency_in = frequency_in
        self.curve_in = curve_in

        self.voltage_out = 0
        self.cv_out.off()
        self.last_tick_at = time.ticks_ms()
        self.change_voltage()

        self.frequency = 0.0
        self.curve_k = 0.0
        self.voltage_out = 0.0

        self.vizualization_samples = []

    def change_voltage(self):
        """Pick a random value between -0.1 and 1.1 and use it as the goal point for the bezier curve

        We intentionally overshoot the [0, 1] range in order to force the occasional effect from the clipping mode.
        """
        self.curve.set_next_value(random.random() * 1.2 - 0.1)

    def update(self, clip_mode=CLIP_MODE_LIMIT):
        now = time.ticks_ms()

        self.curve_k = self.curve_in.percent() * 2 -1  # [-1, 1]
        self.frequency = self.frequency_in.percent() * (self.script.config.MAX_FREQUENCY - self.script.config.MIN_FREQUENCY) + self.script.config.MIN_FREQUENCY

        if self.script.config.AIN_MODE == AIN_MODE_FREQUENCY:
            # increase the frequency according to the voltage on AIN
            self.frequency = self.frequency + ain.read_voltage() / self.script.config.MAX_INPUT_VOLTAGE * (self.script.config.MAX_FREQUENCY - self.script.config.MIN_FREQUENCY) + self.script.config.MIN_FREQUENCY
        elif self.script.config.AIN_MODE == AIN_MODE_CURVE:
            ain_k = ain.read_voltage() / self.script.config.MAX_INPUT_VOLTAGE * 2 -1  # [-1, 1]
            self.curve_k = (self.curve_k + ain_k) / 2

        t = 1000.0/self.frequency  # Hz -> ms

        elapsed_ms = time.ticks_diff(now, self.last_tick_at)
        if elapsed_ms >= t:
            self.change_voltage()
            self.last_tick_at = now
            elapsed_ms = 0

        self.voltage_out = self.curve.value_at(elapsed_ms / t, self.curve_k) * (self.script.config.MAX_VOLTAGE - self.script.config.MIN_VOLTAGE) + self.script.config.MIN_VOLTAGE

        if clip_mode == CLIP_MODE_LIMIT:
            self.voltage_out = self.clip_limit(self.voltage_out)
        elif clip_mode == CLIP_MODE_FOLD:
            self.voltage_out = self.clip_fold(self.voltage_out)
        elif clip_mode == CLIP_MODE_THRU:
            self.voltage_out = self.clip_thru(self.voltage_out)

        self.cv_out.voltage(self.voltage_out)

        self.vizualization_samples.append(int((self.voltage_out - self.script.config.MIN_VOLTAGE) / (self.script.config.MAX_VOLTAGE - self.script.config.MIN_VOLTAGE) * OLED_HEIGHT/3))
        if len(self.vizualization_samples) > OLED_WIDTH // 2:
            self.vizualization_samples.pop(0)

    def clip_limit(self, v):
        if v < self.script.config.MIN_VOLTAGE:
            return self.script.config.MIN_VOLTAGE
        elif v > self.script.config.MAX_VOLTAGE:
            return self.script.config.MAX_VOLTAGE
        else:
            return v

    def clip_fold(self, v):
        if v < self.script.config.MIN_VOLTAGE:
            return self.script.config.MIN_VOLTAGE - v
        elif v > self.script.config.MAX_VOLTAGE:
            return self.script.config.MAX_VOLTAGE + (self.script.config.MAX_VOLTAGE - v)
        else:
            return v

    def clip_thru(self, v):
        if v < self.script.config.MIN_VOLTAGE:
            return self.script.config.MAX_VOLTAGE - (self.script.config.MIN_VOLTAGE - v)
        elif v > self.script.config.MAX_VOLTAGE:
            return self.script.config.MIN_VOLTAGE - (self.script.config.MAX_VOLTAGE - v)
        else:
            return v


class Bezier(EuroPiScript):
    def __init__(self):
        super().__init__()

        cfg = self.load_state_json()

        self.frequency_in = (
            KnobBank.builder(k1)
            .with_unlocked_knob("channel_a")
            .with_locked_knob("channel_b", initial_percentage_value=cfg.get("channel_b_frequency", 0.5))
            .build()
        )

        self.curve_in = (
            KnobBank.builder(k2)
            .with_unlocked_knob("channel_a")
            .with_locked_knob("channel_b", initial_percentage_value=cfg.get("channel_b_curve", 0.5))
            .build()
        )

        # How do we handle clipping?
        self.clip_mode = cfg.get("clip_mode", CLIP_MODE_LIMIT)

        # How do we combine the gate outputs for cv6?
        if self.config.LOGIC_MODE == "and":
            self.gate_logic = LogicalAnd()
        elif self.config.LOGIC_MODE == "or":
            self.gate_logic = LogicalOr()
        elif self.config.LOGIC_MODE == "xor":
            self.gate_logic = LogicalXor()
        elif self.config.LOGIC_MODE == "nand":
            self.gate_logic = LogicalNand()
        elif self.config.LOGIC_MODE == "nor":
            self.gate_logic = LogicalNor()
        elif self.config.LOGIC_MODE == "xnor":
            self.gate_logic = LogicalXnor()
        else:
            raise Exception(f"Unknown logic mode {self.config.LOGIC_MODE}")

        # Are the settings dirty & need saving?
        self.settings_dirty = False

        self.curve_a = OutputChannel(self, self.frequency_in["channel_a"], self.curve_in["channel_a"], cv1)
        self.curve_b = OutputChannel(self, self.frequency_in["channel_b"], self.curve_in["channel_b"], cv2)

        # If DIN receives a signal, force the curves to change target value
        self.force_voltage_change = False

        @din.handler
        def on_din_rise():
            self.force_voltage_change = True

        @b1.handler
        def on_b1_press():
            self.clip_mode = (self.clip_mode + 1) % N_CLIP_MODES
            self.settings_dirty = True
            ssoled.notify_user_interaction()

        @b2.handler
        def on_b2_press():
            self.curve_in.set_current("channel_b")
            self.frequency_in.set_current("channel_b")
            ssoled.notify_user_interaction()

        @b2.handler_falling
        def on_b2_release():
            self.curve_in.set_current("channel_a")
            self.frequency_in.set_current("channel_a")
            self.settings_dirty = True
            ssoled.notify_user_interaction()

    @classmethod
    def config_points(cls):
        """Return the static configuration options for this class
        """
        def restrict_input_voltage(v):
            if v > europi_config.MAX_INPUT_VOLTAGE:
                return europi_config.MAX_INPUT_VOLTAGE
            return v

        return [
            configuration.floatingPoint(
                name="MAX_INPUT_VOLTAGE",
                minimum=0.0,
                maximum=europi_config.MAX_INPUT_VOLTAGE,
                default=restrict_input_voltage(10.0)
            ),
            configuration.floatingPoint(
                name="MIN_VOLTAGE",
                minimum=0.0,
                maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                default=0.0
            ),
            configuration.floatingPoint(
                name="MAX_VOLTAGE",
                minimum=0.0,
                maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                default=europi_config.MAX_OUTPUT_VOLTAGE
            ),
            configuration.floatingPoint(
                name="MIN_FREQUENCY",
                minimum=0.001,
                maximum=10.0,
                default=0.01
            ),
            configuration.floatingPoint(
                name="MAX_FREQUENCY",
                minimum=0.001,
                maximum=10.0,
                default=1.0
            ),
            configuration.choice(
                name="AIN_MODE",
                choices=["frequency", "curve"],
                default="frequency"
            ),
            configuration.choice(
                name="LOGIC_MODE",
                choices=["and", "or", "xor", "nand", "nor", "xnor"],
                default="xor"
            )
        ]

    def save(self):
        """Write the persistent settings file
        """
        cfg = {
            "channel_b_curve": self.curve_in["channel_b"].percent(),
            "channel_b_frequency": self.frequency_in["channel_b"].percent(),
            "clip_mode": self.clip_mode
        }
        self.save_state_json(cfg)
        self.settings_dirty = False

    def draw_graph(self, curve):
        # draw the live graphs
        for i in range(len(curve.vizualization_samples)):
            ssoled.pixel(i+OLED_WIDTH//2, OLED_HEIGHT - 1 - curve.vizualization_samples[i], 1)

    def main(self):
        GATE_DURATION = 0
        HALF_VOLTAGE = (self.config.MIN_VOLTAGE + self.config.MAX_VOLTAGE) / 2

        # Used to detect user interaction with the knobs
        UI_DEADZONE = 0.01
        prev_freq_value = self.frequency_in.current.percent()
        prev_curve_value = self.curve_in.current.percent()

        while True:
            now = time.ticks_ms()

            if self.force_voltage_change:
                self.force_voltage_change = False
                self.curve_a.change_voltage()
                self.curve_b.change_voltage()

            # set the main outputs, calculate the bezier curves
            self.curve_a.update()
            self.curve_b.update()

            # set cv3 to the average of the channels
            cv3.voltage((self.curve_a.voltage_out + self.curve_b.voltage_out) / 2)

            # set the trigger state for cv4
            if time.ticks_diff(now, self.curve_a.last_tick_at) <= GATE_DURATION:
                cv4.on()
                gate_a = True
            else:
                cv4.off()
                gate_a = False

            # set cv5 high/low, depending on the value of channel b
            if self.curve_b.voltage_out >= HALF_VOLTAGE:
                cv5.on()
                gate_b = True
            else:
                cv5.off()
                gate_b = False

            logic_gate_on = self.gate_logic.compare(gate_a, gate_b)
            if logic_gate_on:
                cv6.on()
            else:
                cv6.off()

            # save settings if needed
            if self.settings_dirty:
                self.save()

            # check if we've moved the knobs manually
            # Wake up from the screensaver if we have
            current_freq_value = self.frequency_in.current.percent()
            current_curve_value = self.curve_in.current.percent()
            if abs(current_freq_value - prev_freq_value) >= UI_DEADZONE or abs(current_curve_value - prev_curve_value) >= UI_DEADZONE:
                ssoled.notify_user_interaction()
            prev_freq_value = current_freq_value
            prev_curve_value = current_curve_value

            # UI:
            # +------------------------+
            # | A AA.AAHz  +kA.AA      |
            # | B BB.BBHz  -kB.BB      |
            # | Fold      ...--^-.--^  |
            # +------------------------+
            ssoled.fill(0)

            # Highlight either the first or second row to indicate it's being edited
            channel_a_active = self.curve_in.current == self.curve_in["channel_a"]

            if channel_a_active:
                ssoled.fill_rect(0, 0, OLED_WIDTH, CHAR_HEIGHT+2, 1)
            else:
                ssoled.fill_rect(0, CHAR_HEIGHT+1, OLED_WIDTH, CHAR_HEIGHT+2, 1)

            ssoled.text(f"A {self.curve_a.frequency:0.2f}Hz  {self.curve_a.curve_k:+0.2f}", 1, 1, 0 if channel_a_active else 1)
            ssoled.text(f"B {self.curve_b.frequency:0.2f}Hz  {self.curve_b.curve_k:+0.2f}", 1, CHAR_HEIGHT+2, 1 if channel_a_active else 0)
            ssoled.text(CLIP_MODE_NAMES[self.clip_mode], 1, 2*CHAR_HEIGHT + 4, 1)

            self.draw_graph(self.curve_a)
            self.draw_graph(self.curve_b)

            ssoled.show()

if __name__ == "__main__":
    Bezier().main()



#!/usr/bin/env python3
# Copyright 2023 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Euclidean rhythm generator for the EuroPi

@author Chris Iverach-Brereton <ve4cib@gmail.com>
@year   2023
"""

import random
import time

from europi import *
from europi_script import EuroPiScript

from experimental.euclid import generate_euclidean_pattern
from experimental.screensaver import Screensaver
from experimental.settings_menu import *


class EuclidGenerator:
    """Generates the euclidean rhythm for a single output
    """

    def __init__(self, cv_out, name, steps=1, pulses=0, rotation=0, skip=0):
        """Create a generator that sends its output to the given CV output

        @param cv_out  One of the six output jacks (cv1..cv6)
        @param steps  The initial number of steps (1-32)
        @param pulses  The initial number of pulses (0-32)
        @param rotation  The initial rotation (0-32)
        @param skip  The skip probability (0-1)
        """
        setting_prefix = name.lower()

        self.rotatation = None
        self.pulses = None
        self.steps = None

        self.steps = SettingMenuItem(
            config_point = IntegerConfigPoint(
                f"{setting_prefix}_steps",
                1,
                32,
                steps,
            ),
            prefix = name,
            title = "Steps",
            callback = self.update_steps,
            autoselect_cv = True,
            autoselect_knob = True,
        )

        self.rotation = SettingMenuItem(
            config_point = IntegerConfigPoint(
                f"{setting_prefix}_rotation",
                0,
                32,
                rotation,
            ),
            prefix = name,
            title = "Rotation",
            callback = self.update_rotation,
            autoselect_cv = True,
            autoselect_knob = True,
        )

        self.pulses = SettingMenuItem(
            config_point = IntegerConfigPoint(
                f"{setting_prefix}_pulses",
                0,
                32,
                pulses,
            ),
            prefix = name,
            title = "Pulses",
            callback = self.update_pulses,
            autoselect_cv = True,
            autoselect_knob = True,
        )

        self.skip = SettingMenuItem(
            config_point = IntegerConfigPoint(
                f"{setting_prefix}_skip_prob",
                0,
                100,
                skip,
            ),
            prefix = name,
            title = "Skip %",
            autoselect_cv = True,
            autoselect_knob = True,
        )

        ## The CV output this generator controls
        self.cv = cv_out

        ## The name for this channel
        self.name = name

        ## The current position within the pattern
        self.position = 0

        ## The on/off pattern we generate
        self.pattern = []

        ## Cached copy of the string representation
        #
        #  __str__(self) will do some extra string processing
        #  if this is None; otherwise its value is simply returned
        self.str = None

        # Initialize the pattern
        self.update_steps(self.steps.value, 0, None, None)
        self.regenerate()

    def update_steps(self, new_steps, old_steps, config_point, arg=None):
        """Update the max range of pulses & rotation to match the number of steps
        """
        self.pulses.modify_choices(choices=list(range(new_steps+1)), new_default=new_steps)
        self.rotation.modify_choices(choices=list(range(new_steps+1)), new_default=new_steps)

        self.regenerate()

    def update_pulses(self, new_pulses, old_pulses, config_point, arg=None):
        self.regenerate()

    def update_rotation(self, new_rot, old_rot, config_point, arg=None):
        self.regenerate()

    def __str__(self):
        """Return a string representation of the pattern

        The string consists of 4 characters:
            - ^ current beat, high
            - v current beat, low
            - | high beat
            - . low beat

        e.g. |.|.^|.|.||. is a 7/12 pattern, where the 5th note
        is currently playing
        """

        if self.str is None:
            s = ""
            for i in range(len(self.pattern)):
                if i == self.position:
                    if self.pattern[i] == 0:
                        s = s+"v"
                    else:
                        s = s+"^"
                else:
                    if self.pattern[i] == 0:
                        s = s+"."
                    else:
                        s = s+"|"
            self.str = s
        return self.str

    def regenerate(self):
        """Re-calculate the pattern for this generator

        Call this after changing any of steps, pulses, or rotation to apply
        the changes.

        Changing the pattern will reset the position to zero
        """
        self.position = 0
        self.pattern = generate_euclidean_pattern(self.steps.value, self.pulses.value, self.rotation.value)

        # clear the cached string representation
        self.str = None

    def advance(self):
        """Advance to the next step in the pattern and set the CV output
        """
        # advance the position
        # to ease CPU usage don't do any divisions, just reset to zero
        # if we overflow
        self.position = self.position+1
        if self.position >= len(self.pattern):
            self.position = 0

        if self.steps == 0 or self.pattern[self.position] == 0:
            self.cv.off()
        else:
            if self.skip.value / 100 > random.random():
                self.cv.off()
            else:
                self.cv.on()

        # clear the cached string representation
        self.str = None


class EuclidVisualization(MenuItem):
    """A  menu item for displaying a specific Euclidean channel
    """

    def __init__(self, generator, children=None, parent=None):
        super().__init__(children=children, parent=parent)

        self.generator = generator

    def draw(self, oled=oled):
        pattern_str = str(self.generator)
        oled.text(f"-- {self.generator.name} --", 0, 0)
        if len(pattern_str) > 16:
            pattern_row1 = pattern_str[0:16]
            pattern_row2 = pattern_str[16:]
            oled.text(f"{pattern_row1}", 0, 10)
            oled.text(f"{pattern_row2}", 0, 20)
        else:
            oled.text(f"{pattern_str}", 0, 10)


class EuclideanRhythms(EuroPiScript):
    """Generates 6 different Euclidean rhythms, one per output

    Must be clocked externally into DIN
    """

    def __init__(self):
        super().__init__()

        ## The euclidean pattern generators for each CV output
        #
        #  We pre-load the defaults with some interesting patterns so the script
        #  does _something_ out of the box
        self.generators = [
            EuclidGenerator(cv1, "CV1", 8, 5),
            EuclidGenerator(cv2, "CV2", 16, 7),
            EuclidGenerator(cv3, "CV3", 16, 11),
            EuclidGenerator(cv4, "CV4", 32, 9),
            EuclidGenerator(cv5, "CV5", 32, 15),
            EuclidGenerator(cv6, "CV6", 32, 19)
        ]

        menu_items = []
        for i in range(len(self.generators)):
            menu_items.append(
                EuclidVisualization(
                    self.generators[i],
                    children = [
                        self.generators[i].steps,
                        self.generators[i].pulses,
                        self.generators[i].rotation,
                        self.generators[i].skip,
                    ]
                )
            )

        self.menu = SettingsMenu(
            navigation_knob = k2,
            navigation_button = b2,
            menu_items = menu_items,
            short_press_cb = self.on_menu_short_press,
            long_press_cb = self.on_menu_long_press
        )
        self.menu.load_defaults(self._state_filename)

        # Is the visualization stale (i.e. have we received a pulse and not updated the visualization?)
        self.viz_dirty = True

        self.screensaver = Screensaver()

        self.last_user_interaction_at = time.ticks_ms()

        @din.handler
        def on_rising_clock():
            """Handler for the rising edge of the input clock

            Advance all of the rhythms
            """
            for g in self.generators:
                g.advance()
            self.viz_dirty = True

        @din.handler_falling
        def on_falling_clock():
            """Handler for the falling edge of the input clock

            Turn off all of the CVs so we don't stay on for adjacent pulses
            """
            turn_off_all_cvs()

        @b1.handler
        def on_b1_press():
            """Handler for pressing button 1

            Advance all of the rhythms
            """
            self.last_user_interaction_at = time.ticks_ms()
            for g in self.generators:
                g.advance()
            self.viz_dirty = True

        @b1.handler_falling
        def on_b1_release():
            """Handler for releasing button 1

            Turn off all of the CVs so we don't stay on for adjacent pulses
            """
            self.last_user_interaction_at = time.ticks_ms()
            turn_off_all_cvs()

    def on_menu_long_press(self):
        self.last_user_interaction_at = time.ticks_ms()

    def on_menu_short_press(self):
        self.last_user_interaction_at = time.ticks_ms()

    def main(self):

        # manually check the state of k1 since it's otherwise not used, but should
        # disable the screensaver
        prev_k1 = int(k1.percent() * 100)

        while True:
            now = time.ticks_ms()

            current_k1 = int(k1.percent() * 100)
            if current_k1 != prev_k1:
                self.last_user_interaction_at = now
                prev_k1 = current_k1

            if self.menu.ui_dirty:
                self.last_user_interaction_at = now

            if time.ticks_diff(now, self.last_user_interaction_at) >= self.screensaver.ACTIVATE_TIMEOUT_MS:
                self.last_user_interaction_at = time.ticks_add(now, -self.screensaver.ACTIVATE_TIMEOUT_MS)
                self.screensaver.draw()
            else:
                if self.viz_dirty or self.menu.ui_dirty:
                    self.viz_dirty = False
                    oled.fill(0)
                    self.menu.draw()
                    oled.show()

            if self.menu.settings_dirty:
                self.menu.save(self._state_filename)

if __name__=="__main__":
    EuclideanRhythms().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A clone of Pet Rock by Jonah Senzel.

Tracks the phase of the moon using a realtime clock and generates
pseudo-random gate sequences based on the date & moon phase

The original code is written in C++ and released under the
CC BY-NC-SA 4.0 license (https://creativecommons.org/licenses/by-nc-sa/4.0/legalcode.en)

Original source code is located here: https://github.com/jsenzel1/petrock

This re-implementation is not a perfect 1:1 copy of the original, but attempts
to faithfully recreate the idea behind it for the EuroPi module.

Instead of using colours to differentiate between the moods, this program
displays moods using Tarot and/or playing card suits:
- swords (red)
- cups (blue)
- shields (wands/clubs is the official tarot suit here, but clubs goes with the
  english card suits below, and shields fit with swords, and are found in some
  swiss-german decks of cards) (yellow)
- pentacles (green)

An additional 4 moods were included in the original C++ firmware, but were
deprecated. They are also re-implemented here, but disabled by default:
- hearts
- spades
- diamonds
- clubs
"""

from europi import *
from europi_script import EuroPiScript

import configuration
from framebuf import FrameBuffer, MONO_HLSB
import math
import random
import time

from experimental.a_to_d import AnalogReaderDigitalWrapper
from experimental.experimental_config import RTC_NONE
from experimental.math_extras import rescale
from experimental.random_extras import shuffle
from experimental.rtc import *


class Algo:
    """
    Generic algorithm for generating the gate sequences

    All pseudo-random algorithms inherit from this class
    """

    CHANNEL_A = 1
    CHANNEL_B = 2

    def __init__(self, channel, weekday, cycle, continuity, algorithm, mood_name):
        """
        Child constructors must call this first

        Child constructors must initialize self.sequence by appending {0, 1} values to it

        @param channel  1 for channel A, 2 for channel B
        @param weekday  The current weekday 1-7 (M-Su)
        @param cycle  The current moon phase: 0-7 (new to waning crescent)
        @param continuity  A random value shared between both A and B channels: 0-100
        @param algorithm  The name of the generator algorithm (used for logging only)
        @param mood_name  The name of the mood symbol/suit (used for logging only)
        """
        self.algorithm_name = algorithm
        self.mood_name = mood_name

        self.channel = channel

        if channel == self.CHANNEL_A:
            self.gate_out = cv1
            self.inv_out = cv2
            self.eos_out = cv3
        else:
            self.gate_out = cv4
            self.inv_out = cv5
            self.eos_out = cv6

        self.weekday = weekday
        self.cycle = cycle
        self.continuity = continuity

        self.sequence = []
        self.index = 0

        self.state_dirty = False

    def sanitize_sequence(self, sequence=None):
        """
        Ensure that the sequence is neither all-1 nor all-0

        If either is true, flip a random element

        This should be called after generating the sequence in the child class constructors

        @param sequence  If none, self.sequence is sanitized
        """
        if sequence is None:
            sequence = self.sequence

        empty = True
        full = True
        for n in sequence:
            if n == 0:
                full = False
            else:
                empty = False

        # flip an item if the pattern is wholly uniform
        if empty or full:
            sequence[random.randint(0, len(sequence) -1)] = (sequence[random.randint(0, len(sequence) -1)] + 1) % 2

    def __str__(self):
        return f"{self.sequence}"

    def __eq__(self, other):
        """
        Return True if both sequences are identical
        """
        if len(self.sequence) == len(other.sequence):
            for i in range(len(self.sequence)):
                if self.sequence[i] != other.sequence[i]:
                    return False
            return True
        return False

    @staticmethod
    def map(x, in_min, in_max, out_min, out_max):
        # treat the output as inclusive
        out_max = out_max + 1
        return int((x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min)

    def tick(self):
        """
        Advance the sequence
        """
        self.index = (self.index + 1 ) % len(self.sequence)
        self.state_dirty = True

    def set_outputs(self):
        """
        Set the outputs high/low as needed
        """
        if self.sequence[self.index]:
            self.gate_out.on()
            self.inv_out.off()
        else:
            self.gate_out.off()
            self.inv_out.on()

        if self.index == len(self.sequence) - 1:
            self.eos_out.on()

        self.state_dirty = False

    def outputs_off(self):
        self.gate_out.off()
        self.inv_out.off()
        self.eos_out.off()


class AlgoPlain(Algo):
    """
    A straight-forward random-fill

    We choose a length based on the moon phase and then fill it with on/off signals
    using a coin-toss.

    This _could_ result in all-on or all-off patterns, but this is generally unlikely

    This corresponds to the red mood in the original firmware
    """

    # swords
    mood_graphics = bytearray(b'\x00\x00\x00\x1f\x00\x00\x00!\x00\x00\x00A\x00\x00\x00\x81\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x04\x04\x00\x00\x08\x08\x00\x00\x10\x10\x00\x00  \x00\x00@@\x00\x00\x80\x80\x00\x01\x01\x00\x00\x02\x02\x00\x04\x04\x04\x00\x04\x08\x08\x00\x06\x10\x10\x00\x07  \x00\x03\xc0@\x00\x01\xc0\x80\x00\x00\xe1\x00\x00\x00r\x00\x00\x00\xfc\x00\x00\x01\xdc\x00\x00\x03\x8e\x00\x00\x07\x07\x00\x00~\x03\xc0\x00|\x00\x00\x00|\x00\x00\x00|\x00\x00\x00|\x00\x00\x00\x00\x00\x00\x00')

    def __init__(self, channel, weekday, cycle, continuity):
        super().__init__(channel, weekday, cycle, continuity, "plain", "swords")

        seqmax = 0

        if cycle == MoonPhase.NEW_MOON:
            seqmax = random.randint(5, 7)
        elif cycle == MoonPhase.WAXING_CRESCENT or cycle == MoonPhase.WANING_CRESCENT:
            seqmax = random.randint(4, 16)
        elif cycle == MoonPhase.FIRST_QUARTER or cycle == MoonPhase.THIRD_QUARTER:
            seqmax = Algo.map(continuity, 0, 100, 6, 12)
            seqmax = seqmax * self.channel  # channel B is twice as long as A
        elif cycle == MoonPhase.WAXING_GIBBOUS:
            seqmax = 12
        elif cycle == MoonPhase.WANING_GIBBOUS:
            seqmax = 16
        else:
            seqmax = 16

        # From Jonah S:
        # Randomly populate rhythm
        # this may seems weird/cheesy, but as I note in the manual - I decided to focus on
        # one and only one translated elements for the moon cycle, which is the length
        # relationship of A and B - I found through practice that the random population of
        # steps actually produces great results, the key is how many steps you use, and the
        # relationship between the 2 step lengths. The interesting difference is comparing
        # for example a pair of 8 step rhythms, vs a 7 step rhythm, and a 15 step rhythm
        # being played against each other - this is the "meta movement" of the rhythmic
        # flavor, in every algo/mood
        for i in range(seqmax):
            self.sequence.append(random.randint(0, 1))

        self.sanitize_sequence()


class AlgoReich(Algo):
    """
    Choose a random pattern length pased on the moon phase and our continuity variable
    and fill the pattern with on/off using a fixed density

    This corresponds to the blue mood in the original firmware
    """

    # cups
    mood_graphics = bytearray(b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x7f\xff\xff\xfe\x7f\xff\xff\xfe\x7f\xff\xff\xfe?\xff\xff\xfc?\xff\xff\xfc\x1f\xff\xff\xf8\x0f\xff\xff\xf0\x07\xff\xff\xe0\x03\xff\xff\xc0\x01\xff\xff\x80\x00\x7f\xfe\x00\x00\x0f\xf0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x0f\xf0\x00\x00\x7f\xfe\x00\x01\xff\xff\x80\x03\xff\xff\xc0')

    def __init__(self, channel, weekday, cycle, continuity):
        super().__init__(channel, weekday, cycle, continuity, "reich", "cups")

        if cycle == MoonPhase.NEW_MOON:
            seqmax = random.randint(3, 5)
        elif cycle == MoonPhase.WAXING_CRESCENT or cycle == MoonPhase.WAXING_CRESCENT:
            if channel == Algo.CHANNEL_A:
                seqmax = Algo.map(continuity, 0, 100, 3, 8)
            else:
                a = Algo.map(continuity, 0, 100, 3, 8)
                b = 0
                while b == 0 or b == a or b == a*2 or b*2 == a:
                    b = random.randint(3, 8)

                seqmax = b
        elif cycle == MoonPhase.FIRST_QUARTER or cycle == MoonPhase.THIRD_QUARTER:
            seqmax = Algo.map(continuity, 0, 100, 5, 9)
            seqmax = seqmax * channel  # B is double A
        elif cycle == MoonPhase.WAXING_GIBBOUS or cycle == MoonPhase.WANING_GIBBOUS:
            seqmax = Algo.map(continuity, 0, 100, 4, 8)
        else:
            seqmax = 8

        seqDensity=50
        for i in range(seqmax):
            if random.randint(0, 99) < seqDensity:
                self.sequence.append(1)
            else:
                self.sequence.append(0)

        self.sanitize_sequence()


class AlgoSparse(Algo):
    """
    Chooses a fixed length based on moon phase and fills the pattern with a low-density
    on/off pattern (10% on)

    This corresponds to the yellow mood in the original firmware
    """

    # shields
    mood_graphics = bytearray(b'\xff\xff\xff\xff\x9f\xff\xff\xff\x8f\xff\xe0?\x87\xff\xf0\x7f\x83\xff\xb8\xef\xc1\xff\x98\xcf\xe0\xff\x80\x0f\xf0\x7f\x80\x0f\xf8?\x80\x0f\xfc\x1f\x98\xcf\xfe\x0f\xb8\xef\xff\x07\xf0\x7f\xff\x83\xe0?\xff\xc1\xff\xff\xff\xe0\xff\xff\xff\xf0\x7f\xff\xff\xf8?\xff\x7f\xfc\x1f\xfe?\xfe\x0f\xfc\x1f\xff\x07\xf8\x0f\xff\x83\xf0\x07\xff\xc1\xe0\x03\xff\xe0\xc0\x01\xff\xf0\x80\x00\xff\xf9\x00\x00\x7f\xfe\x00\x00?\xfc\x00\x00\x1f\xf8\x00\x00\x0f\xf0\x00\x00\x07\xe0\x00\x00\x03\xc0\x00\x00\x01\x80\x00')

    def __init__(self, channel, weekday, cycle, continuity):
        super().__init__(channel, weekday, cycle, continuity, "sparse", "shields")

        if cycle == MoonPhase.NEW_MOON:
            seqmax = random.randint(10, 19)
        elif cycle == MoonPhase.WAXING_CRESCENT or cycle == MoonPhase.WANING_CRESCENT:
            seqmax = random.randint(15, 30)
        elif cycle == MoonPhase.FIRST_QUARTER:
            if channel == Algo.CHANNEL_A:
                seqmax = 32
            else:
                seqmax = 64
        elif cycle == MoonPhase.THIRD_QUARTER:
            if channel == Algo.CHANNEL_A:
                seqmax = 24
            else:
                seqmax = 48
        elif cycle == MoonPhase.WAXING_GIBBOUS:
            seqmax = 32
        elif cycle == MoonPhase.WANING_GIBBOUS:
            seqmax = 24
        else:
            seqmax = 64

        densityPercent = 10

        for i in range(seqmax):
            self.sequence.append(0)

        seedStepInd = random.randint(0, seqmax - 1)
        self.sequence[seedStepInd] = 1

        for i in range(seqmax):
            if random.randint(0, 99) < densityPercent:
                self.sequence[i] = 1

        self.sanitize_sequence()


class AlgoVari(Algo):
    """
    Generates two sub-sequences A & B, repeating each a fixed number of times

    e.g. if sequence a is 0011 and sequence b is 1010, with 2 repeats the final pattern
    is 0011 0011 1010 1010

    This corresponds to the green mood in the original firmware
    """

    # pentacles
    mood_graphics = bytearray(b'\x00\x07\xe0\x00\x009\x9c\x00\x00\xc1\x83\x00\x01\x01\x80\x80\x02\x02@@\x04\x02@ \x08\x02@\x10\x10\x04 \x08 \x04 \x04 \x04 \x04@\x08\x10\x02\x7f\xff\xff\xfeP\x08\x10\n\x88\x10\x08\x11\x84\x10\x08!\x83\x10\x08\xc1\x80\xa0\x05\x01\x80`\x06\x01\x800\x0c\x01@H\x12\x02@Fb\x02@A\x82\x02 \x81\x81\x04 \x86a\x04\x10\x88\x11\x08\t0\x0c\x90\x05@\x02\xa0\x03\x80\x01\xc0\x01\x00\x00\x80\x00\xc0\x03\x00\x008\x1c\x00\x00\x07\xe0\x00')

    def __init__(self, channel, weekday, cycle, continuity):
        super().__init__(channel, weekday, cycle, continuity, "vari", "pentacles")

        if cycle == MoonPhase.NEW_MOON:
            seqmax = random.randint(3, 19)
            repeats = 3
        elif cycle == MoonPhase.WAXING_CRESCENT or cycle == MoonPhase.WANING_CRESCENT:
            seqmax = random.randint(8, 12)
            repeats = random.randint(3, 6)
        elif cycle == MoonPhase.FIRST_QUARTER or cycle == MoonPhase.THIRD_QUARTER:
            seqmax = 8
            if channel == Algo.CHANNEL_A:
                repeats = 4
            else:
                repeats = 8
        elif cycle == MoonPhase.WAXING_GIBBOUS or cycle == MoonPhase.WANING_GIBBOUS:
            seqmax = 16
            repeats = 4
        else:
            seqmax = 12
            repeats = 3

        seq_a = []
        seq_b = []
        for i in range(seqmax):
            r = random.randint(0, 1)
            seq_a.append(r)
            seq_b.append(r)

        for i in range(seqmax-1, -1, -1):
            j = random.randint(0, i)

            tmp = seq_b[i]
            seq_b[i] = seq_b[j]
            seq_b[j] = i

        # the whole sequence is r * [seq_a] + r * [seq_b]
        for r in range(repeats):
            for n in seq_a:
                self.sequence.append(n)
        for r in range(repeats):
            for n in seq_b:
                self.sequence.append(n)

        self.sanitize_sequence()


class AlgoBlocks(Algo):
    """
    Builds the sequence by randomly choosing N pre-defined pattern blocks

    One of the unimplemented algorithms in the original firmware
    """

    # hearts
    mood_graphics = bytearray(b'\x07\xe0\x07\xe0\x1f\xf8\x1f\xf8?\xfc?\xfc\x7f\xfe\x7f\xfe\x7f\xfe\x7f\xfe\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x7f\xff\xff\xfe?\xff\xff\xfc\x1f\xff\xff\xf8\x1f\xff\xff\xf8\x0f\xff\xff\xf0\x07\xff\xff\xe0\x07\xff\xff\xe0\x03\xff\xff\xc0\x01\xff\xff\x80\x01\xff\xff\x80\x00\xff\xff\x00\x00\x7f\xfe\x00\x00\x7f\xfe\x00\x00?\xfc\x00\x00\x1f\xf8\x00\x00\x1f\xf8\x00\x00\x0f\xf0\x00\x00\x07\xe0\x00\x00\x07\xe0\x00\x00\x03\xc0\x00\x00\x01\x80\x00')

    blocks = [
        [0, 0, 0, 0],
        [1, 0, 0, 0],
        [1, 0, 1, 0],
        [1, 0, 0, 1],
        [1, 1, 1, 0],
        [1, 0, 1, 1],
        [0, 0, 1, 0],
        [1, 1, 1, 1],
    ]

    def __init__(self, channel, weekday, cycle, continuity):
        super().__init__(channel, weekday, cycle, continuity, "blocks", "hearts")

        # This is wholly original
        # the original, deprecated code has a to-do here
        if cycle == MoonPhase.NEW_MOON:
            numblocks = 2
        elif cycle == MoonPhase.WAXING_CRESCENT or cycle == MoonPhase.WANING_CRESCENT:
            numblocks = random.randint(3, 4)
        elif cycle == MoonPhase.FIRST_QUARTER or cycle == MoonPhase.THIRD_QUARTER:
            numblocks = random.randint(2, 3)
            numblocks = numblocks * self.channel  # B is double A
        elif cycle == MoonPhase.WAXING_GIBBOUS or cycle == MoonPhase.WANING_GIBBOUS:
            numblocks = random.randint(4, 5)
        else:
            numblocks = 6

        for i in range(numblocks):
            block = AlgoBlocks.blocks[random.randint(0, len(AlgoBlocks.blocks) - 1)]
            for n in block:
                self.sequence.append(n)

        self.sanitize_sequence()


class AlgoCulture(Algo):
    """
    Chooses a pre-programmed rhythm based on the weekday and adds a random number
    of coin-toss steps to the end of it

    The back half of the rhythms is zero, resulting in some interesting hand-offs between
    the main and inverted outputs.

    One of the unimplemented algorithms in the original firmware
    """

    # spades
    mood_graphics = bytearray(b'\x00\x01\x80\x00\x00\x03\xc0\x00\x00\x07\xe0\x00\x00\x0f\xf0\x00\x00\x1f\xf8\x00\x00?\xfc\x00\x00\x7f\xfe\x00\x00\xff\xff\x00\x01\xff\xff\x80\x03\xff\xff\xc0\x07\xff\xff\xe0\x0f\xff\xff\xf0\x1f\xff\xff\xf8?\xff\xff\xfc\x7f\xff\xff\xfe\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x7f\xff\xff\xfe\x7f\xff\xff\xfe?\xfd\xbf\xfc\x1f\xf9\x9f\xf8\x07\xe1\x87\xe0\x00\x01\x80\x00\x00\x01\x80\x00\x00\x01\x80\x00\x00\x07\xe0\x00\x00\x1f\xf8\x00\x00?\xfc\x00')

    rhythms = [
        [1,0,0,1,1,0,1,0,1,0,0,1,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
        [1,0,0,1,0,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
        [1,0,0,1,0,0,1,0,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
        [1,0,1,0,1,0,1,0,1,0,0,1,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
        [1,1,1,1,1,0,0,1,1,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
        [1,1,0,1,1,0,1,0,1,1,0,1,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
        [1,0,0,1,0,0,1,0,0,0,1,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
        [1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    ]

    def __init__(self, channel, weekday, cycle, continuity):
        super().__init__(channel, weekday, cycle, continuity, "culture", "spades")

        # add or remove some steps based on the moon phase
        if cycle == MoonPhase.NEW_MOON:
            extra_steps = random.randint(-16, -8)
        elif cycle == MoonPhase.WAXING_CRESCENT or cycle == MoonPhase.WANING_CRESCENT:
            extra_steps = random.randint(0, 4)
        elif cycle == MoonPhase.FIRST_QUARTER or cycle == MoonPhase.THIRD_QUARTER:
            extra_steps = random.randint(-8, 8)
        elif cycle == MoonPhase.WAXING_GIBBOUS or cycle == MoonPhase.WANING_GIBBOUS:
            extra_steps = random.randint(-4, 12)
        else:
            extra_steps = random.randint(-4, 4)

        rhythm = AlgoCulture.rhythms[weekday]
        for i in range(len(rhythm) + extra_steps):
            if i < len(rhythm):
                self.sequence.append(rhythm[i])
            else:
                self.sequence.append(0)

        # backfill the second half-ish of the rhythm with randomness
        for i in range(15, len(self.sequence)):
            self.sequence[i] = random.randint(0, 1)

        self.sanitize_sequence()


class AlgoOver(Algo):
    """
    Generates two sub-sequences and overwrites parts of the main sequence with each
    after each complete cycle

    One of the unimplemented algorithms in the original firmware
    """

    # diamonds
    mood_graphics = bytearray(b'\x00\x01\x80\x00\x00\x01\x80\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x07\xe0\x00\x00\x07\xe0\x00\x00\x0f\xf0\x00\x00\x0f\xf0\x00\x00\x1f\xf8\x00\x00\x1f\xf8\x00\x00?\xfc\x00\x00?\xfc\x00\x00\x7f\xfe\x00\x00\xff\xff\x00\x03\xff\xff\xc0\x0f\xff\xff\xf0\x0f\xff\xff\xf0\x03\xff\xff\xc0\x00\xff\xff\x00\x00\x7f\xfe\x00\x00?\xfc\x00\x00?\xfc\x00\x00\x1f\xf8\x00\x00\x1f\xf8\x00\x00\x0f\xf0\x00\x00\x0f\xf0\x00\x00\x07\xe0\x00\x00\x07\xe0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x01\x80\x00\x00\x01\x80\x00')

    def __init__(self, channel, weekday, cycle, continuity):
        super().__init__(channel, weekday, cycle, continuity, "over", "diamonds")

        densityPercent = 50

        # same lengths as AlgoPlain
        if cycle == MoonPhase.NEW_MOON:
            seqmax = random.randint(5, 7)
        elif cycle == MoonPhase.WAXING_CRESCENT or cycle == MoonPhase.WANING_CRESCENT:
            seqmax = random.randint(4, 16)
        elif cycle == MoonPhase.FIRST_QUARTER or cycle == MoonPhase.THIRD_QUARTER:
            seqmax = Algo.map(continuity, 0, 100, 6, 12)
            seqmax = seqmax * self.channel  # channel B is twice as long as A
        elif cycle == MoonPhase.WAXING_GIBBOUS:
            seqmax = 12
        elif cycle == MoonPhase.WANING_GIBBOUS:
            seqmax = 16
        else:
            seqmax = 16

        self.seq1 = []
        self.seq2 = []

        for i in range(seqmax):
            if random.randint(0, 99) < densityPercent:
                self.seq1.append(1)
            else:
                self.seq1.append(0)

            if random.randint(0, 99) < densityPercent:
                self.seq2.append(1)
            else:
                self.seq2.append(0)


        self.swaps = list(range(seqmax))
        shuffle(self.swaps)
        self.switch_index = 0
        self.swap = True

        self.sanitize_sequence(self.seq1)
        self.sanitize_sequence(self.seq2)

        for n in self.seq1:
            self.sequence.append(n)

    def tick(self):
        super().tick()

        # overwrite steps betwen seq1 and seq2
        if self.index == 0:
            overwrite_index = self.swaps[self.switch_index]
            self.switch_index += 1

            if self.switch_index > len(self.sequence) - 1:
                self.switch_index = 0
                self.swap = not self.swap

            if not self.swap:
                self.sequence[overwrite_index] = self.seq2[overwrite_index]
            else:
                self.sequence[overwrite_index] = self.seq1[overwrite_index]


class AlgoWonk(Algo):
    """
    Generates a fixed-length pattern with a random density based on the weekday. We then
    "wonk" a number of steps based on the moon phase

    Wonking means we choose a random on-step, turn it off, and then turn the step before
    or after on.

    One of the unimplemented algorithms in the original firmware
    """

    # clubs
    mood_graphics = bytearray(b'\x00\x07\xe0\x00\x00\x0f\xf0\x00\x00\x1f\xf8\x00\x00?\xfc\x00\x00\x7f\xfe\x00\x00\x7f\xfe\x00\x00\x7f\xfe\x00\x00\x7f\xfe\x00\x00\x7f\xfe\x00\x00\x7f\xfe\x00\x0f\xff\xff\xf0\x1f\xff\xff\xf8?\xff\xff\xfc\x7f\xff\xff\xfe\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x7f\xfb\xdf\xfe?\xf3\xcf\xfc\x1f\xe3\xc7\xf8\x0f\xc3\xc3\xf0\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x03\xc0\x00\x00\x07\xe0\x00\x00\x0f\xf0\x00\x00\x1f\xf8\x00\x00?\xfc\x00\x00\x7f\xfe\x00')

    def __init__(self, channel, weekday, cycle, continuity):
        super().__init__(channel, weekday, cycle, continuity, "wonk", "clubs")

        densityPercent = Algo.map(weekday, 1, 7, 30, 60)

        seqmax = 32

        for i in range(seqmax):
            self.sequence.append(0)


        # ensure there is at least 1 filled step
        self.sequence[random.randint(0, seqmax - 1)] = 1

        steps_placed = 0
        for i in range(0, seqmax, 4):
            if random.randint(0, 99) < densityPercent:
                self.sequence[i] = 1
                steps_placed += 1

        steps_to_wonk = Algo.map(cycle, 0, 7, 1, steps_placed)
        steps_wonked = 0
        while steps_wonked < steps_to_wonk:
            chosen_step = random.randint(0, seqmax - 1)

            if self.sequence[chosen_step] == 1:
                steps_wonked += 1
                self.sequence[chosen_step] = 0

                r = random.randint(0, 1)
                if chosen_step == 0:
                    self.sequence[chosen_step + 1] = 1
                elif chosen_step == seqmax - 1:
                    self.sequence[chosen_step - 1] = 1
                elif r == 0:
                    self.sequence[chosen_step + 1] = 1
                else:
                    self.sequence[chosen_step - 1] = 1

        self.sanitize_sequence()


class MoonPhase:
    """
    Calculates the current moon phase
    """

    NEW_MOON = 0
    WAXING_CRESCENT = 1
    FIRST_QUARTER = 2
    WAXING_GIBBOUS = 3
    FULL_MOON = 4
    WANING_GIBBOUS = 5
    THIRD_QUARTER = 6
    WANING_CRESCENT = 7

    moon_phase_images = [
        bytearray(b'\x00\x07\xc0\x00\x000\x0c\x00\x00\x80\x01\x00\x03\x00\x00\xc0\x04\x00\x00 \x08\x00\x00\x10\x00\x00\x00\x00\x10\x00\x00\x08 \x00\x00\x04\x00\x00\x00\x00@\x00\x00\x02@\x00\x00\x02\x00\x00\x00\x00\x80\x00\x00\x01\x80\x00\x00\x01\x80\x00\x00\x01\x80\x00\x00\x01\x80\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00@\x00\x00\x02@\x00\x00\x02\x00\x00\x00\x00 \x00\x00\x04\x10\x00\x00\x08\x10\x00\x00\x08\x08\x00\x00\x10\x04\x00\x00 \x01\x00\x00\x80\x00\x80\x01\x00\x000\x0c\x00\x00\x07\xc0\x00'),
        bytearray(b'\x00\x0b\xe0\x00\x000\xfe\x00\x00\x80\x1f\x00\x01\x00\x0f\xc0\x04\x00\x03\xe0\x00\x00\x01\xf0\x00\x00\x00\xf8\x10\x00\x00\xf8 \x00\x00|\x00\x00\x00>@\x00\x00>@\x00\x00>\x00\x00\x00\x1e\x00\x00\x00\x1f\x80\x00\x00\x1f\x00\x00\x00\x1f\x80\x00\x00\x1f\x00\x00\x00\x1f\x00\x00\x00\x1f\x00\x00\x00\x1e@\x00\x00>@\x00\x00>\x00\x00\x00< \x00\x00|\x10\x00\x00\xf8\x00\x00\x00\xf8\x08\x00\x01\xf0\x00\x00\x03\xe0\x01\x00\x0f\xc0\x00\x80\x1f\x00\x000\xfc\x00\x00\x0b\xe0\x00'),
        bytearray(b'\x00\x02\xe0\x00\x000\xfe\x00\x00\x80\xff\x00\x01\x00\xff\xc0\x04\x00\xff\xe0\x00\x00\xff\xf0\x00\x00\xff\xf8\x10\x00\xff\xf8 \x00\xff\xfc\x00\x00\xff\xfe@\x00\xff\xfe@\x00\xff\xfe\x00\x00\xff\xfe\x00\x00\xff\xff\x80\x00\xff\xff\x00\x00\xff\xff\x80\x00\xff\xff\x00\x00\xff\xff\x00\x00\xff\xff\x00\x00\xff\xfe@\x00\xff\xfe@\x00\xff\xfe\x00\x00\xff\xfc \x00\xff\xfc\x10\x00\xff\xf8\x00\x00\xff\xf8\x08\x00\xff\xf0\x00\x00\xff\xe0\x01\x00\xff\xc0\x00\x80\xff\x00\x000\xfc\x00\x00\x02\xe0\x00'),
        bytearray(b'\x00\x02\x90\x00\x000\xfc\x00\x00\x83\xff\x00\x01\x07\xff\xc0\x04\x0f\xff\xe0\x00\x1f\xff\xf0\x00?\xff\xf8\x10\x7f\xff\xf8 \x7f\xff\xfc\x00\xff\xff\xfe@\xff\xff\xfe@\xff\xff\xfe\x01\xff\xff\xfe\x01\xff\xff\xff\x81\xff\xff\xff\x01\xff\xff\xff\x81\xff\xff\xff\x01\xff\xff\xff\x01\xff\xff\xff\x01\xff\xff\xfe@\xff\xff\xfe@\xff\xff\xfe\x00\xff\xff\xfc \x7f\xff\xfc\x10\x7f\xff\xf8\x00?\xff\xf8\x08\x1f\xff\xf0\x00\x0f\xff\xe0\x01\x07\xff\xc0\x00\x83\xff\x00\x000\xfc\x00\x00\x02\x90\x00'),
        bytearray(b'\x00\x07\xc0\x00\x00?\xfc\x00\x00\xff\xff\x00\x03\xff\xff\xc0\x07\xff\xff\xe0\x0f\xff\xff\xf0\x0f\xff\xff\xf0\x1f\xff\xff\xf8?\xff\xff\xfc?\xff\xff\xfc\x7f\xff\xff\xfe\x7f\xff\xff\xfe\x7f\xff\xff\xfe\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x7f\xff\xff\xfe\x7f\xff\xff\xfe\x7f\xff\xff\xfe\x7f\xff\xff\xfe?\xff\xff\xfc?\xff\xff\xfc\x1f\xff\xff\xf8\x1f\xff\xff\xf8\x0f\xff\xff\xf0\x07\xff\xff\xe0\x01\xff\xff\x80\x00\xff\xff\x00\x00?\xfc\x00\x00\x07\xc0\x00'),
        bytearray(b'\x00\t\x80\x00\x00\x7f\x0c\x00\x00\xff\xc1\x00\x03\xff\xe0\x80\x07\xff\xf0 \x0f\xff\xf8\x00\x1f\xff\xfc\x00\x1f\xff\xfe\x08?\xff\xfe\x04\x7f\xff\xff\x00\x7f\xff\xff\x02\x7f\xff\xff\x02\xff\xff\xff\x80\x7f\xff\xff\x80\xff\xff\xff\x81\xff\xff\xff\x81\xff\xff\xff\x80\xff\xff\xff\x81\xff\xff\xff\x80\x7f\xff\xff\x80\x7f\xff\xff\x02\x7f\xff\xff\x02?\xff\xff\x00?\xff\xfe\x04\x1f\xff\xfe\x08\x1f\xff\xfc\x00\x0f\xff\xf8\x00\x07\xff\xf0\x00\x03\xff\xe0\x80\x00\xff\xc1\x00\x00\x7f\x0c\x00\x00\t\x80\x00'),
        bytearray(b'\x00\x0b\x80\x00\x00\x7f\x0c\x00\x00\xff\x01\x00\x03\xff\x00\x80\x07\xff\x00 \x0f\xff\x00\x00\x1f\xff\x00\x00\x1f\xff\x00\x08?\xff\x00\x04\x7f\xff\x00\x00\x7f\xff\x00\x02\x7f\xff\x00\x02\xff\xff\x00\x00\x7f\xff\x00\x00\xff\xff\x00\x01\xff\xff\x00\x01\xff\xff\x00\x00\xff\xff\x00\x01\xff\xff\x00\x00\x7f\xff\x00\x00\x7f\xff\x00\x02\x7f\xff\x00\x02?\xff\x00\x00?\xff\x00\x04\x1f\xff\x00\x08\x1f\xff\x00\x00\x0f\xff\x00\x00\x07\xff\x00\x00\x03\xff\x00\x80\x00\xff\x01\x00\x00\x7f\x0c\x00\x00\x0b\x80\x00'),
        bytearray(b'\x00\x0b\xe0\x00\x00\x7f\x0c\x00\x00\xf8\x01\x00\x03\xf0\x00\x80\x07\xc0\x00 \x0f\x80\x00\x00\x1f\x00\x00\x00\x1f\x00\x00\x08>\x00\x00\x04|\x00\x00\x00|\x00\x00\x02|\x00\x00\x02\xf8\x00\x00\x00x\x00\x00\x00\xf8\x00\x00\x01\xf8\x00\x00\x01\xf8\x00\x00\x00\xf8\x00\x00\x01\xf8\x00\x00\x00x\x00\x00\x00|\x00\x00\x02|\x00\x00\x02<\x00\x00\x00>\x00\x00\x04\x1f\x00\x00\x08\x1f\x00\x00\x00\x0f\x80\x00\x00\x07\xc0\x00\x00\x03\xf0\x00\x80\x00\xf8\x01\x00\x00\x7f\x0c\x00\x00\x0b\xe0\x00'),
    ]

    @staticmethod
    def calculate_days_since_new_moon(date):
        """
        Calculate the number of days since a known full moon

        @see https://www.subsystems.us/uploads/9/8/9/4/98948044/moonphase.pdf

        @param date  The current UTC DateTime
        """
        if date.year < 2000:
            raise ValueError(f"Date out of range; check your RTC")

        y = date.year
        m = date.month
        d = date.day

        if m == Month.JANUARY or m == Month.FEBRUARY:
            y = y - 1
            m = m + 12

        a = math.floor(y/100)
        b = math.floor(a/4)
        c = 2 - a + b
        e = math.floor(365.25 * (y + 4716))
        f = math.floor(30.6001 * (m + 1))
        jd = c + d + e + f - 1524.5

        days_since_new_moon = jd - 2451549.5

        return days_since_new_moon

    @staticmethod
    def calculate_phase(date):
        """
        Calculate the current moon phase

        @param date  The current UTC DateTime
        """
        days_since_new_moon = MoonPhase.calculate_days_since_new_moon(date)

        yesterday_new_moons = (days_since_new_moon - 1) / 29.53
        today_new_moons = days_since_new_moon / 29.53
        tomorrow_new_moons = (days_since_new_moon + 1) / 29.53

        # we always want 1 day assigned to new, first quarter, full, and third quarter
        # so use yesterday, today, and tomorrow as a 3-day window
        # if tomorrow is on one side of the curve and yesterday was the other, treat today
        # as the "special" phase
        yesterday_fraction = yesterday_new_moons % 1
        today_fraction = today_new_moons % 1
        tomorrow_fraction = tomorrow_new_moons % 1

        # first cases are for the special 1-day events
        # check if we're in the transition area, and then decide if today is actually closest
        # to the event
        if yesterday_fraction > 0.75 and tomorrow_fraction < 0.25:
            if MoonPhase.closest_to_fraction(0.0, yesterday_fraction, today_fraction, tomorrow_fraction) == 0:
                return MoonPhase.NEW_MOON
            elif today_fraction > 0.75:
                return MoonPhase.WANING_CRESCENT
            else:
                return MoonPhase.WAXING_CRESCENT
        elif yesterday_fraction < 0.25 and tomorrow_fraction > 0.25:
            if MoonPhase.closest_to_fraction(0.25, yesterday_fraction, today_fraction, tomorrow_fraction) == 0:
                return MoonPhase.FIRST_QUARTER
            elif today_fraction < 0.25:
                return MoonPhase.WAXING_CRESCENT
            else:
                return MoonPhase.WAXING_GIBBOUS
        elif yesterday_fraction < 0.5 and tomorrow_fraction > 0.5:
            if MoonPhase.closest_to_fraction(0.5, yesterday_fraction, today_fraction, tomorrow_fraction) == 0:
                return MoonPhase.FULL_MOON
            elif today_fraction < 0.5:
                return MoonPhase.WAXING_GIBBOUS
            else:
                return MoonPhase.WANING_GIBBOUS
        elif yesterday_fraction < 0.75 and tomorrow_fraction > 0.75:
            if MoonPhase.closest_to_fraction(0.75, yesterday_fraction, today_fraction, tomorrow_fraction) == 0:
                return MoonPhase.THIRD_QUARTER
            elif today_fraction < 0.75:
                return MoonPhase.WANING_GIBBOUS
            else:
                return MoonPhase.WANING_CRESCENT
        elif today_fraction == 0.0:
            return MoonPhase.NEW_MOON
        elif today_fraction < 0.25:
            return MoonPhase.WAXING_CRESCENT
        elif today_fraction == 0.25:
            return MoonPhase.FIRST_QUARTER
        elif today_fraction < 0.5:
            return MoonPhase.WAXING_GIBBOUS
        elif today_fraction == 0.5:
            return MoonPhase.FULL_MOON
        elif today_fraction < 0.75:
            return MoonPhase.WANING_GIBBOUS
        elif today_fraction == 0.75:
            return MoonPhase.THIRD_QUARTER
        else:
            return MoonPhase.WANING_CRESCENT

    @staticmethod
    def closest_to_fraction(fraction, yesterday, today, tomorrow):
        yesterday = abs(fraction - yesterday)
        today = abs(fraction - today)
        tomorrow = abs(fraction - tomorrow)

        if yesterday < tomorrow and yesterday < today:  # yesterday is closest
            return -1
        elif today < yesterday and today < tomorrow:  # today is closest
            return 0
        else:
            return 1  # tomorrow is closest


class Mood:
    """
    The algorithm's "mood"

    Mood is one of 4 colours, which rotates every moon cycle
    """

    available_moods = [
        AlgoPlain,
        AlgoReich,
        AlgoSparse,
        AlgoVari
    ]

    @staticmethod
    def set_moods(moodstring):
        if moodstring == "all":
            Mood.available_moods = [
                AlgoPlain,
                AlgoReich,
                AlgoSparse,
                AlgoVari,
                AlgoBlocks,
                AlgoCulture,
                AlgoOver,
                AlgoWonk,
            ]
        elif moodstring == "alternate":
            Mood.available_moods = [
                AlgoBlocks,
                AlgoCulture,
                AlgoOver,
                AlgoWonk,
            ]
        else:  # "classic"
            Mood.available_moods = [
                AlgoPlain,
                AlgoReich,
                AlgoSparse,
                AlgoVari,
            ]

    @staticmethod
    def mood_algorithm(date):
        """
        Get the algorithm for the current mood

        @param date  The current UTC DateTime
        """
        days_since_new_moon = MoonPhase.calculate_days_since_new_moon(date)
        cycles = math.floor(days_since_new_moon / 29.53)

        return Mood.available_moods[cycles % len(Mood.available_moods)]


class IntervalTimer:
    """
    Uses ticks_ms and ticks_diff to fire a callback at fixed-ish intervals
    """

    MIN_INTERVAL = 10
    MAX_INTERVAL = 500

    def __init__(self, speed_knob, rise_cb=lambda: None, fall_cb=lambda: None):
        self.interval_ms = 0
        self.last_tick_at = time.ticks_ms()

        self.speed_knob = speed_knob

        self.rise_callback = rise_cb
        self.fall_callback = fall_cb

        self.next_rise = True

        self.update_interval()

    def update_interval(self):
        DEADZONE = 0.1
        p = self.speed_knob.percent()
        if p <= DEADZONE:
            # disable the timer for the first 10% of travel so we have an easy-off
            # for external clocking
            self.interval_ms = 0
        else:
            p = 1.0 - rescale(p, DEADZONE, 1, 0, 1)
            self.interval_ms = round(rescale(p, 0, 1, self.MIN_INTERVAL, self.MAX_INTERVAL))

    def tick(self):
        # kick out immediately if the timer is off
        if self.interval_ms <= 0:
            return

        now = time.ticks_ms()
        if time.ticks_diff(now, self.last_tick_at) >= self.interval_ms:
            self.last_tick_at = now

            if self.next_rise:
                self.rise_callback()
                self.next_rise = False
            else:
                self.fall_callback()
                self.next_rise = True


class NoClockError(Exception):
    """
    A custom exception we can raise if the RTC is not configured correctly
    """
    def __init__(self, message):
        super().__init__(message)


class PetRock(EuroPiScript):

    def __init__(self):
        super().__init__()

        # validate the clock and raise a NoClockError if
        # we are using the default no-RTC implementation
        if experimental_config.RTC_IMPLEMENTATION == RTC_NONE:
            raise NoClockError("No clock configured")

        Mood.set_moods(self.config.MOODS)

        self.seed_offset = 1
        self.generate_sequences(clock.utcnow())

        self.din2 = AnalogReaderDigitalWrapper(
            ain,
            cb_rising = self.on_channel_b_trigger,
            cb_falling = self.on_channel_b_fall
        )
        b2.handler(self.on_channel_b_trigger)
        b2.handler_falling(self.on_channel_b_fall)
        self.timer_b = IntervalTimer(
            k2,
            rise_cb=self.on_channel_b_trigger,
            fall_cb=self.on_channel_b_fall
        )

        din.handler(self.on_channel_a_trigger)
        din.handler_falling(self.on_channel_a_fall)
        b1.handler(self.on_channel_a_trigger)
        b1.handler_falling(self.on_channel_a_fall)
        self.timer_a = IntervalTimer(
            k1,
            rise_cb=self.on_channel_a_trigger,
            fall_cb=self.on_channel_a_fall
        )

    @classmethod
    def config_points(cls):
        return [
            # What moods does the user want to use?
            # - classic: the original 4 moods from the hardware Pet Rock
            # - alternate: the 4 deprecated moods that weren't implemented
            # - all: all 8 possible moods
            configuration.choice(
                "MOODS",
                choices=[
                    "classic",
                    "alternate",
                    "all"
                ],
                default="classic",
            ),
        ]

    def generate_sequences(self, now):
        """
        Regenerate the day's rhythms

        @param now  The current UTC
        """
        if now.weekday is None:
            now.weekday = 0

        local_weekday = (now + local_timezone).weekday

        continuity = random.randint(0, 99)
        cycle = MoonPhase.calculate_phase(now)
        today_seed = now.day + now.month + now.year + self.seed_offset
        random.seed(today_seed)

        self.sequence_a = Mood.mood_algorithm(now)(Algo.CHANNEL_A, local_weekday, cycle, continuity)
        self.sequence_b = Mood.mood_algorithm(now)(Algo.CHANNEL_B, local_weekday, cycle, continuity)

        self.last_generation_at = clock.localnow()

        # print a YAML-like block with the sequences for debugging
        print(f"# Generated sequences at {self.last_generation_at} ({local_timezone})")
        print("parameters:")
        print(f"  continuity: {continuity}")
        print(f"  date: '{now} UTC'")
        print(f"  moon_phase: {cycle}")
        print(f"  today_seed: {today_seed}")
        print(f"  weekday: {local_weekday}")
        print("sequences:")
        print(f"  algorithm: {self.sequence_a.algorithm_name}")
        print(f"  mood: {self.sequence_a.mood_name}")
        print(f"  seq_a: {self.sequence_a}")
        print(f"  seq_b: {self.sequence_b}")

    def on_channel_a_trigger(self):
        self.sequence_a.tick()

    def on_channel_a_fall(self):
        self.sequence_a.outputs_off()

    def on_channel_b_trigger(self):
        self.sequence_b.tick()

    def on_channel_b_fall(self):
        self.sequence_b.outputs_off()

    def draw(self, utc_time):
        oled.fill(0)

        local_time = utc_time + local_timezone
        if local_time.weekday:
            oled.text(Weekday.NAME[local_time.weekday][0:3].upper(), OLED_WIDTH - CHAR_WIDTH * 3, 0, 1)

        oled.text(f"{local_time.hour:02}:{local_time.minute:02}", OLED_WIDTH - CHAR_WIDTH * 5, OLED_HEIGHT - CHAR_HEIGHT, 1)

        moon_phase = MoonPhase.calculate_phase(utc_time)
        moon_img = FrameBuffer(MoonPhase.moon_phase_images[moon_phase], 32, 32, MONO_HLSB)
        oled.blit(moon_img, 0, 0)

        mood_img = FrameBuffer(self.sequence_a.mood_graphics, 32, 32, MONO_HLSB)
        oled.blit(mood_img, 40, 0)

        oled.show()

    def run_test(self):
        self.draw(clock.utcnow())
        last_draw_at = clock.utcnow()

        fake_date = clock.utcnow()


        yesterday_moon_phase = -1
        while True:
            self.din2.update()

            self.timer_a.update_interval()
            self.timer_b.update_interval()

            self.timer_a.tick()
            self.timer_b.tick()

            local_time = clock.utcnow()

            ui_dirty = local_time.minute != last_draw_at.minute

            if ui_dirty:
                fake_date.minute = local_time.minute
                fake_date.hour = local_time.hour
                fake_date.day = fake_date.day + 1
                fake_date.weekday = fake_date.weekday + 1
                if fake_date.weekday == 8:
                    fake_date.weekday = 1
                if fake_date.day > fake_date.days_in_month:
                    fake_date.day = 1
                    fake_date.month = fake_date.month + 1

                    if fake_date.month == 13:
                        fake_date.month = 1
                        fake_date.year += 1

                self.generate_sequences(fake_date)
                self.sequence_a.state_dirty = True
                self.sequence_b.state_dirty = True

                # check that we don't have two consecutive special-phase days
                today_moon_phase = MoonPhase.calculate_phase(fake_date)
                if (
                    today_moon_phase == MoonPhase.NEW_MOON or
                    today_moon_phase == MoonPhase.FIRST_QUARTER or
                    today_moon_phase == MoonPhase.FULL_MOON or
                    today_moon_phase == MoonPhase.THIRD_QUARTER
                ):
                    if today_moon_phase == yesterday_moon_phase:
                        print(f"WARNING: two consecutive {today_moon_phase}-phases!")

                # check that the two sequences are different
                if self.sequence_a == self.sequence_b:
                    print("WARNING: identical sequences generated!")



                self.draw(fake_date)
                last_draw_at = local_time

            if self.sequence_a.state_dirty:
                self.sequence_a.set_outputs()

            if self.sequence_b.state_dirty:
                self.sequence_b.set_outputs()

    def main(self):
        self.draw(clock.utcnow())
        last_draw_at = clock.localnow()

        while True:
            self.din2.update()

            self.timer_a.update_interval()
            self.timer_b.update_interval()

            self.timer_a.tick()
            self.timer_b.tick()

            local_time = clock.localnow()

            ui_dirty = local_time.minute != last_draw_at.minute

            # if the day has rolled over, generate new sequences and mark them as dirty
            # so we'll continue playing
            if local_time.day != self.last_generation_at.day:
                self.generate_sequences(clock.utcnow())
                self.sequence_a.state_dirty = True
                self.sequence_b.state_dirty = True

                ui_dirty = True

            if self.sequence_a.state_dirty:
                self.sequence_a.set_outputs()

            if self.sequence_b.state_dirty:
                self.sequence_b.set_outputs()

            if ui_dirty:
                self.draw(clock.utcnow())
                last_draw_at = local_time


if __name__ == "__main__":
    PetRock().main()



# Copyright 2025 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from europi_script import EuroPiScript

from experimental.a_to_d import AnalogReaderDigitalWrapper
from experimental.math_extras import rescale

clock_input = din
reset_input = AnalogReaderDigitalWrapper(ain)
advance_button = b1
reset_button = b2

advance_output = cv1
end_of_sequence_output = cv2

sequence_length_knob = k1
step_size_knob = k2

# Duration of triggers sent to DFAM to advance
# based on experimentation, 1ms is sufficient and doesn't produce
# audio artefacts when burst-advancing
TRIGGER_DURATION = 0.001

# What length of sequences can we output?
# We allow up to DFAM's max length of 8
MIN_SEQUENCE_LEN = 1
MAX_SEQUENCE_LEN = 8

# How many steps does DFAM advance with each trigger
# This is all modular, so setting to 7 will advance 1 space backwards
MIN_STEP = 1
MAX_STEP = 7

# DFAM has an 8-step sequencer
DFAM_SEQUENCER_LENGTH = 8

# The size of the circles we draw on the UI
CIRCLE_SIZE = 4

# Screen colours
BLACK = 0
WHITE = 1

# Circle fill constants
YES_FILL = -1
NO_FILL = 0


class DfamController(EuroPiScript):
    def __init__(self):
        super().__init__()

        self.reset_request = False
        self.advance_request = False

        # the current step within our internal sequencer
        self.current_step = 0

        # the maximum number of steps in the sequence
        self.max_steps = MAX_SEQUENCE_LEN

        # the number of advances we output per input step
        self.step_size = MIN_STEP

        # the current step on DFAM's sequencer
        self.dfam_sync_counter = 0

        clock_input.handler(self.request_advance)
        advance_button.handler(self.request_advance)

        reset_input.handler(self.request_reset)
        reset_button.handler(self.request_reset)

    def request_reset(self):
        self.reset_request = True

    def request_advance(self):
        self.advance_request = True

    def reset(self, force_trigger=False):
        pulses = (DFAM_SEQUENCER_LENGTH - self.dfam_sync_counter) % DFAM_SEQUENCER_LENGTH
        if force_trigger and pulses == 0:
            pulses = DFAM_SEQUENCER_LENGTH
        self.current_step = 0
        self.dfam_sync_counter = 0
        for _ in range(pulses):
            advance_output.on()
            time.sleep(TRIGGER_DURATION)
            advance_output.off()
            time.sleep(TRIGGER_DURATION)

    def advance(self):
        for _ in range(self.step_size):
            advance_output.on()
            time.sleep(TRIGGER_DURATION)
            advance_output.off()
            time.sleep(TRIGGER_DURATION)
        self.dfam_sync_counter = (self.dfam_sync_counter + self.step_size) % DFAM_SEQUENCER_LENGTH

    def main(self):
        render_needed = True

        while True:
            reset_input.update()  # a-to-d wrapper, so we need to poll it!

            new_steps = int(rescale(sequence_length_knob.percent(), 0, 1, MIN_SEQUENCE_LEN, MAX_SEQUENCE_LEN))
            if new_steps != self.max_steps:
                render_needed = True
                self.max_steps = new_steps

            new_size = int(rescale(step_size_knob.percent(), 0, 1, MIN_STEP, MAX_STEP))
            if new_size != self.step_size:
                render_needed = True
                self.step_size = new_size

            if self.reset_request:
                render_needed = True
                self.reset_request = False
                self.reset()

            if self.advance_request:
                render_needed = True
                self.advance_request = False
                self.current_step += 1
                if self.current_step >= self.max_steps:
                    if self.max_steps == 1:
                        self.advance()

                    end_of_sequence_output.on()
                    self.reset(force_trigger=True)
                else:
                    end_of_sequence_output.off()
                    self.advance()

            if render_needed:
                render_needed = False
                oled.fill(0)
                oled.centre_text(f"{self.current_step + 1}/{self.max_steps}\nx{self.step_size}\n", auto_show=False, clear_first=False)

                for i in range(DFAM_SEQUENCER_LENGTH):
                    # the 0th LED is actually the last one, so the pattern should be shifted 1 to the left
                    if i == (self.dfam_sync_counter - 1) % DFAM_SEQUENCER_LENGTH:
                        fill = YES_FILL
                    else:
                        fill = NO_FILL
                    oled.ellipse(
                        i * OLED_WIDTH // DFAM_SEQUENCER_LENGTH + CIRCLE_SIZE,
                        OLED_HEIGHT - CIRCLE_SIZE - 1,
                        CIRCLE_SIZE,
                        CIRCLE_SIZE,
                        WHITE,
                        fill
                    )

                oled.show()


if __name__ == "__main__":
    DfamController().main()
